package timemlmerger;

import Comparators.AscINT_lid_Comparator;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import nlp_files.XMLFile;
import utils_bk.*;
import xml_parsers.XmlAttribs;

/**
 *
 * @author Hector Llorens
 */
public class TimeML_Merger {

    public static void merge(ArrayList<File[]> annotations, double[] weights, File output_folder, double min_et, String link_threshold, String links_merge_option) {
        try {
            // use annotations[0] as key-guide
            File[] guide = annotations.get(0);
            // for each file
            for (int i = 0; i < guide.length; i++) {
                System.out.println("Merging ..." + guide[i].getName());
                HashMap<String, String>[] timex_map = new HashMap[annotations.size()]; // original-normalized id
                HashMap<String, String>[] event_map = new HashMap[annotations.size()]; // original-normalized id
                HashMap<String, String>[] mk_map = new HashMap[annotations.size()]; // original-normalized id
                HashMap<String, String>[] event_attr_from_mk_map = new HashMap[annotations.size()]; // original-id-mk attributes
                HashMap<String, String>[] timex_extent_value_map = new HashMap[annotations.size()]; // normalized-id-extent-value-map
                HashMap<String, String>[] event_class_map = new HashMap[annotations.size()]; // normalized-id-extent-value-map
                HashMap<String, Double>[] weighted_entities = new HashMap[annotations.size()];
                HashMap<String, ArrayList<String>>[] event_mk_index = new HashMap[annotations.size()]; // normalied id - related mks
                HashMap<String, String> normmk_normevent_map = new HashMap<String, String>();
                HashMap<String, Double> weighted_links = new HashMap<String, Double>();
                HashMap<String, String> links_support_map = new HashMap<String, String>();
                HashMap<String, String> links_support_map_names = new HashMap<String, String>();
                HashSet<String> merged_entities = new HashSet<String>();
                ArrayList<String> xmlFileString = new ArrayList();
                ArrayList<String[]> tokenFileStringArr = new ArrayList();
                int last_eid = 1;
                int open_event = 0;
                int last_tid = 1;
                int[] open_timex = new int[annotations.size()];
                int[] last_tid_local = new int[annotations.size()];
                String last_text_blanks = "";
                // for each annotation
                for (int a = 0; a < annotations.size(); a++) {
                    File annot = annotations.get(a)[i];
                    File ftdir = new File(annot.getCanonicalPath() + "-data");
                    if (ftdir.exists()) {
                        FileUtils.deleteRecursively(ftdir);
                    }
                    if (!ftdir.mkdirs()) {  // mkdirs creates many parent dirs if needed
                        throw new Exception("Directory not created...");
                    }
                    File workingfile = new File(ftdir + File.separator + annot.getName());
                    FileUtils.copyFileUtil(annot, workingfile);
                    XMLFile xmlfile = new XMLFile();
                    xmlfile.loadFile(workingfile);
                    String plainfile = xmlfile.toPlain(); // only <text>
                    String output = Tokenizer.run(plainfile);
                    output = addSpacing(output, plainfile);
                    output = merge_tok_n_tml(output, workingfile.getCanonicalPath());
                    last_text_blanks = get_last_text_blanks(plainfile);
                    //System.out.println("last blanks:"+last_text_blanks+".");
                    tokenFileStringArr.add(FileUtils.readFileAsString(output, "UTF-8").split("\\n"));
                    xmlFileString.add(FileUtils.readFileAsString(workingfile.getCanonicalPath(), "UTF-8"));
                    event_map[a] = new HashMap<String, String>();
                    timex_map[a] = new HashMap<String, String>();
                    mk_map[a] = new HashMap<String, String>();
                    event_attr_from_mk_map[a] = new HashMap<String, String>();
                    timex_extent_value_map[a] = new HashMap<String, String>();
                    event_class_map[a] = new HashMap<String, String>();
                    event_mk_index[a] = new HashMap<String, ArrayList<String>>();
                    weighted_entities[a] = new HashMap<String, Double>();
                    open_timex[a] = 0;
                    last_tid_local[a] = 0;
                }
                for (int linen = 0; linen < tokenFileStringArr.get(0).length; linen++) {
                    open_event = 0; // restart in every iteration (multi-token are not considered)
                    String last_token = null; // check all files are equal
                    for (int a = 0; a < annotations.size(); a++) {
                        String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                        if (last_token == null) {
                            last_token = pipesarr[0];
                        } else {
                            if (pipesarr[0].equals(last_token)) {
                                last_token = pipesarr[0];
                            } else {
                                throw new Exception("Tokens doesn't match at line: " + linen + " (" + last_token + ")");
                            }
                        }
                        HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                        if (pipesarr[2].matches("B-EVENT")) {
                            if (open_event == 0) {
                                open_event = last_eid;
                                last_eid++;
                            }
                            event_map[a].put(attribs.get("eid"), "e" + open_event);
                            event_class_map[a].put("e" + open_event, attribs.get("class"));
                        }
                        // for timex maintain open_timex[] first
                        if (pipesarr[2].equals("O") || pipesarr[2].equals("B-TIMEX3")) {
                            open_timex[a] = 0; // restart
                        }
                    }
                    for (int a = 0; a < annotations.size(); a++) {
                        String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                        HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                        if (pipesarr[2].equals("B-TIMEX3")) {
                            open_timex[a] = last_tid; // new by default
                            // try to match with still open timexes
                            for (int at = 0; at < annotations.size(); at++) {
                                // open and not used in other annotations
                                if (a != at && open_timex[at] != 0 && last_tid_local[a] < open_timex[at]) {
                                    open_timex[a] = open_timex[at];
                                    break;
                                }
                            }
                            last_tid_local[a] = open_timex[a];
                            //new (not paired)
                            if (open_timex[a] == last_tid) {
                                last_tid++;
                            }
                            timex_map[a].put(attribs.get("tid"), "t" + open_timex[a]);
                            timex_extent_value_map[a].put("t" + open_timex[a], linen + "|" + linen + "|" + attribs.get("value"));
                        }
                        // update extent of if open timex
                        if (pipesarr[2].matches("I-TIMEX3")) {
                            timex_extent_value_map[a].put("t" + open_timex[a], timex_extent_value_map[a].get("t" + open_timex[a]).replaceFirst("\\|[0-9]+\\|", "|" + linen + "|"));
                        }
                    }
                }
                // normalize entities given a map
                for (int a = 0; a < annotations.size(); a++) {
                    String annotname = annotations.get(a)[i].getCanonicalPath().substring(0, annotations.get(a)[i].getCanonicalPath().lastIndexOf(File.separator));
                    annotname = annotname.substring(annotname.lastIndexOf(File.separator) + 1);
                    String tmp = xmlFileString.get(a);
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    DocumentBuilder db = dbf.newDocumentBuilder();
                    Document doc = db.parse(annotations.get(a)[i]);
                    doc.getDocumentElement().normalize();
                    String dctid = null;
                    Element dct = ((Element) ((NodeList) ((Element) doc.getElementsByTagName("DCT").item(0)).getElementsByTagName("TIMEX3")).item(0));
                    if (dct != null) {
                        dctid = dct.getAttribute("tid");
                        timex_map[a].put(dctid, "t0");
                        // normalize to t0
                        tmp = tmp.replaceAll("(<DCT><TIMEX3[^>]*tid=\")" + dct.getAttribute("tid") + "(\"[^>]*>)", "$1t0$2");
                    }
                    NodeList text = doc.getElementsByTagName("TEXT");
                    if (text.getLength() > 1) {
                        throw new Exception("More than one TEXT tag found.");
                    }
                    Element TextElmnt = (Element) text.item(0); // If not ELEMENT NODE will throw exception
                    // normalize events (entity by entity in order) ---- // deprecated solution: e1=e3 -- e3=e1 map problem. //for(String e:event_map[a].keySet()){tmp=tmp.replaceAll("(eid|eventID)=\""+e+"\"", event_map[a].get(e));}
                    NodeList current_node = TextElmnt.getElementsByTagName("EVENT");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        tmp = tmp.replaceAll("(<EVENT[^>]*eid=\")" + element.getAttribute("eid") + "(\"[^>]*>)", "$1n" + event_map[a].get(element.getAttribute("eid")) + "$2");
                    }
                    tmp = tmp.replaceAll("(<EVENT[^>]*eid=\")n", "$1");
                    // normalize timexes
                    current_node = TextElmnt.getElementsByTagName("TIMEX3");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        if (timex_map[a].get(element.getAttribute("tid")) != null) {
                            tmp = tmp.replaceAll("(<TIMEX3[^>]*tid=\")" + element.getAttribute("tid") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("tid")) + "$2");
                            if (element.hasAttribute("anchorTimeID")) {
                                if (timex_map[a].get(element.getAttribute("anchorTimeID")) != null) {
                                    tmp = tmp.replaceAll("(<TIMEX3[^>]*anchorTimeID=\")" + element.getAttribute("anchorTimeID") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("anchorTimeID")) + "$2");
                                } else {
                                    tmp = tmp.replaceAll("anchorTimeID=\"" + element.getAttribute("anchorTimeID") + "\"", "");
                                }
                            }
                            if (element.hasAttribute("beginPoint")) {
                                if (timex_map[a].get(element.getAttribute("beginPoint")) != null) {
                                    tmp = tmp.replaceAll("(<TIMEX3[^>]*beginPoint=\")" + element.getAttribute("beginPoint") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("beginPoint")) + "$2");
                                } else {
                                    tmp = tmp.replaceAll("beginPoint=\"" + element.getAttribute("beginPoint") + "\"", "");
                                }
                            }
                            if (element.hasAttribute("endPoint")) {
                                if (timex_map[a].get(element.getAttribute("endPoint")) != null) {
                                    tmp = tmp.replaceAll("(<TIMEX3[^>]*endPoint=\")" + element.getAttribute("endPoint") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("endPoint")) + "$2");
                                } else {
                                    tmp = tmp.replaceAll("endPoint=\"" + element.getAttribute("endPoint") + "\"", "");
                                }
                            }
                        } else {
                            tmp = tmp.replaceAll("<TIMEX3[^>]*tid=\"" + element.getAttribute("tid") + "\"[^>]*>([^<]*)</TIMEX3>", "$1");
                            tmp = tmp.replaceAll("anchorTimeID=\"" + element.getAttribute("tid") + "\"", "");
                            tmp = tmp.replaceAll("beginPoint=\"" + element.getAttribute("tid") + "\"", "");
                            tmp = tmp.replaceAll("endPoint=\"" + element.getAttribute("tid") + "\"", "");
                            tmp = tmp.replaceAll("<[TSA]LINK[^>]*=\"" + element.getAttribute("tid") + "\"[^>]*>", "");
                        }
                    }
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*tid=\")n", "$1");
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*anchorTimeID=\")n", "$1");
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*beginPoint=\")n", "$1");
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*endPoint=\")n", "$1");
                    // normalize and map makeinstances
                    HashSet<String> mks_new_ids = new HashSet<String>();
                    current_node = null;
                    current_node = doc.getElementsByTagName("MAKEINSTANCE");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        String mapped_event = event_map[a].get(element.getAttribute("eventID"));
                        if (!event_mk_index[a].containsKey(mapped_event)) {
                            // add attribs for the first
                            String event_attr_from_mk = "";
                            if (element.hasAttribute("pos")) {
                                event_attr_from_mk += "pos=\"" + element.getAttribute("pos") + "\" ";
                            }
                            if (element.hasAttribute("tense")) {
                                event_attr_from_mk += "tense=\"" + element.getAttribute("tense") + "\" ";
                            }
                            if (element.hasAttribute("aspect")) {
                                event_attr_from_mk += "aspect=\"" + element.getAttribute("aspect") + "\" ";
                            }
                            if (element.hasAttribute("polarity")) {
                                event_attr_from_mk += "polarity=\"" + element.getAttribute("polarity") + "\" ";
                            }
                            if (element.hasAttribute("modality")) {
                                event_attr_from_mk += "modality=\"" + element.getAttribute("modality") + "\" ";
                            }
                            event_attr_from_mk_map[a].put(element.getAttribute("eventID"), event_attr_from_mk);
                            ArrayList<String> mks = new ArrayList<String>();
                            if (mks_new_ids.contains(mapped_event.replaceFirst("e", "ei"))) {
                                throw new Exception("Duplicated eiid in mks: " + mapped_event.replaceFirst("e", "ei") + "  " + annotations.get(a)[i]);
                            }
                            normmk_normevent_map.put(mapped_event.replaceFirst("e", "ei"), mapped_event);
                            mks_new_ids.add(mapped_event.replaceFirst("e", "ei"));
                            mks.add(mapped_event.replaceFirst("e", "ei"));
                            event_mk_index[a].put(mapped_event, mks);
                            mk_map[a].put(element.getAttribute("eiid"), mapped_event.replaceFirst("e", "ei"));
                            tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*>)", "$1n" + mapped_event + "$2");
                            tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*>)", "$1n" + mapped_event + "$2");
                            tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\")" + element.getAttribute("eiid") + "(\"[^>]*>)", "$1n" + mapped_event.replaceFirst("e", "ei") + "$2");
                        } else {
                            if (mks_new_ids.contains(mapped_event.replaceFirst("e", "ei"))) {
                                int num = 1000000 + Integer.parseInt(mapped_event.substring(1));
                                while (mks_new_ids.contains("ei" + num)) {
                                    num++;
                                }
                                mks_new_ids.add("ei" + num);
                                normmk_normevent_map.put("ei" + num, mapped_event);
                                event_mk_index[a].get(mapped_event).add("ei" + num);
                                mk_map[a].put(element.getAttribute("eiid"), "ei" + num);
                                tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*>)", "$1n" + mapped_event + "$2");
                                tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*>)", "$1n" + mapped_event + "$2");
                                tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\")" + element.getAttribute("eiid") + "(\"[^>]*>)", "$1n" + "ei" + num + "$2");
                            } else {
                                throw new Exception("Extrange... mkinstance");
                            }
                        }
                    }
                    tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\")n", "$1");
                    tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eventID=\")n", "$1");
                    // normalize links
                    current_node = doc.getElementsByTagName("TLINK");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        String relType = element.getAttribute("relType");
                        if (relType.matches("(DURING|DURING_INV|IDENTITY)")) {
                            relType = "SIMULTANEOUS";
                        }
                        String entity1 = null;
                        String entity2 = null;
                        // event-event
                        if (element.hasAttribute("eventInstanceID") && element.hasAttribute("relatedToEventInstance")) {
                            if (!mk_map[a].containsKey(element.getAttribute("eventInstanceID"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("eventInstanceID"));
                            }
                            if (!mk_map[a].containsKey(element.getAttribute("relatedToEventInstance"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("relatedToEventInstance"));
                            }
                            entity1 = mk_map[a].get(element.getAttribute("eventInstanceID"));
                            entity2 = mk_map[a].get(element.getAttribute("relatedToEventInstance"));
                            // Order by id (for normalization)
                            if (Integer.parseInt(entity1.substring(2)) > Integer.parseInt(entity2.substring(2))) {
                                entity1 = entity2;
                                entity2 = mk_map[a].get(element.getAttribute("eventInstanceID"));
                                relType = reverseRelationCategory(relType);
                            }
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*eventInstanceID=\")" + element.getAttribute("eventInstanceID") + "(\"[^>]*>)", "$1" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relatedToEventInstance=\")" + element.getAttribute("relatedToEventInstance") + "(\"[^>]*>)", "$1" + entity2 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relType=\")" + element.getAttribute("relType") + "(\"[^>]*>)", "$1" + relType + "$2");
                        }
                        // event-time
                        if (element.hasAttribute("eventInstanceID") && element.hasAttribute("relatedToTime")) {
                            if (!mk_map[a].containsKey(element.getAttribute("eventInstanceID"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("eventInstanceID"));
                            }
                            if (!timex_map[a].containsKey(element.getAttribute("relatedToTime"))) {
                                throw new Exception("Timex not found it file: " + element.getAttribute("relatedToTime"));
                            }
                            entity1 = mk_map[a].get(element.getAttribute("eventInstanceID"));
                            entity2 = timex_map[a].get(element.getAttribute("relatedToTime"));
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*eventInstanceID=\")" + element.getAttribute("eventInstanceID") + "(\"[^>]*>)", "$1" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relatedToTime=\")" + element.getAttribute("relatedToTime") + "(\"[^>]*>)", "$1" + entity2 + "$2");
                        }
                        if (element.hasAttribute("timeID") && element.hasAttribute("relatedToEventInstance")) {
                            if (!mk_map[a].containsKey(element.getAttribute("relatedToEventInstance"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("relatedToEventInstance"));
                            }
                            if (!timex_map[a].containsKey(element.getAttribute("timeID"))) {
                                throw new Exception("Timex not found it file: " + element.getAttribute("timeID"));
                            }
                            entity1 = mk_map[a].get(element.getAttribute("relatedToEventInstance"));
                            entity2 = timex_map[a].get(element.getAttribute("timeID"));
                            relType = reverseRelationCategory(relType);
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*)timeID=\"" + element.getAttribute("timeID") + "(\"[^>]*>)", "$1eventInstanceID=\"" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*)relatedToEventInstance=\"" + element.getAttribute("relatedToEventInstance") + "(\"[^>]*>)", "$1relatedToTime=\"" + entity2 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relType=\")" + element.getAttribute("relType") + "(\"[^>]*>)", "$1" + relType + "$2");
                        }
                        // time-time
                        if (element.hasAttribute("timeID") && element.hasAttribute("relatedToTime")) {
                            if (!timex_map[a].containsKey(element.getAttribute("relatedToTime"))) {
                                throw new Exception("Timex instance not found it file: " + element.getAttribute("relatedToTime"));
                            }
                            if (!timex_map[a].containsKey(element.getAttribute("timeID"))) {
                                throw new Exception("Timex not found it file: " + element.getAttribute("timeID"));
                            }
                            entity1 = timex_map[a].get(element.getAttribute("timeID"));
                            entity2 = timex_map[a].get(element.getAttribute("relatedToTime"));
                            // Order by id (for normalization)
                            if (Integer.parseInt(entity1.substring(1)) > Integer.parseInt(entity2.substring(1))) {
                                entity1 = entity2;
                                entity2 = timex_map[a].get(element.getAttribute("timeID"));
                                relType = reverseRelationCategory(relType);
                            }
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*timeID=\")" + element.getAttribute("timeID") + "(\"[^>]*>)", "$1" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relatedToTime=\")" + element.getAttribute("relatedToTime") + "(\"[^>]*>)", "$1" + entity2 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relType=\")" + element.getAttribute("relType") + "(\"[^>]*>)", "$1" + relType + "$2");
                        }
                        // save and calculate link-weight (#triples*100+#doubles)
                        // with weights #triples*(100+weriht_i/10+#doubles*(1+weriht_i/10))
                        // for ties sort higher weighted annotations first (weight is normalized to sum 100)
                        // save and calculate link-weight (#triples*(n*10)+#doubles*n)
                        // with weights #triples*(n*10)+#doubles*n+(system weights)
                        // for ties sort higher weighted annotations first (weight is normalized to sum 100)
                        // upadte links support map
                        if (links_support_map.containsKey(entity1 + "|" + entity2 + "|" + relType)) {
                            links_support_map.put(entity1 + "|" + entity2 + "|" + relType, links_support_map.get(entity1 + "|" + entity2 + "|" + relType) + ";" + a);
                            links_support_map_names.put(entity1 + "|" + entity2 + "|" + relType, links_support_map_names.get(entity1 + "|" + entity2 + "|" + relType) + ";" + annotname);
                        } else {
                            // add new triplet
                            links_support_map.put(entity1 + "|" + entity2 + "|" + relType, "" + a);
                            links_support_map_names.put(entity1 + "|" + entity2 + "|" + relType, "" + annotname);
                        }
                        // check triplet
                        if (weighted_links.containsKey(entity1 + "|" + entity2 + "|" + relType)) {
                            weighted_links.put(entity1 + "|" + entity2 + "|" + relType, weighted_links.get(entity1 + "|" + entity2 + "|" + relType) + (annotations.size() + 1) * weights[a]);
                        } else {
                            // add new triplet
                            weighted_links.put(entity1 + "|" + entity2 + "|" + relType, (annotations.size() + 1) * weights[a]);
                        }
                        //check matching doublet
                        String doublets = weighted_links.keySet().toString();
                        Pattern pattern = Pattern.compile("" + entity1 + "\\|" + entity2 + "\\|[^,\\]]+");
                        Matcher matcher = pattern.matcher(doublets);
                        if (matcher.find()) {
                            for (int dblt = 0; dblt <= matcher.groupCount(); dblt++) {
                                String matchdbl = matcher.group(dblt);
                                weighted_links.put(matchdbl, weighted_links.get(matchdbl) + weights[a]);
                            }
                        }
                    }
                    // For safety ALINKS and SLINKs are removed, also rlinks
                    tmp = tmp.replaceAll("<[ASR]LINK.*", "");
                    // save normalized annotation
                    FileUtils.writeFileFromString(tmp, annotations.get(a)[i].getCanonicalPath() + "-data" + File.separator + annotations.get(a)[i].getName() + ".normalized");
                }
                // weighted links //System.out.println(weighted_links);
                String weightedlinks = saveWeightedLinks(weighted_links, links_support_map_names, annotations.get(0)[i], output_folder);
                String mergedlinks = LinkMerger.run(weightedlinks,links_merge_option);
                HashMap<String, String> makeinstances = new HashMap<String, String>();
                HashMap<String, String> links = new HashMap<String, String>();
                obtainInfoFromMergedLinks(mergedlinks, weights, links_support_map, normmk_normevent_map, weighted_entities, makeinstances, links, link_threshold);
                if (System.getProperty("DEBUG") == null || System.getProperty("DEBUG").equalsIgnoreCase("false")) {
                    (new File(weightedlinks)).delete();
                    (new File(mergedlinks)).delete();
                }
                String line;
                boolean textfound = false;
                String header = "";
                String footer = "";
                BufferedReader modelReader = new BufferedReader(new FileReader(guide[i]));
                try {
                    // read out header
                    while ((line = modelReader.readLine()) != null) {
                        if (line.length() > 0) {
                            if (line.matches(".*<TEXT>.*")) {
                                textfound = true;
                                break;
                            }
                        }
                        header += line + "\n";
                    }
                    if (!textfound) {
                        throw new Exception("Premature end of file (" + guide[i] + ") <TEXT> not found.");
                    }
                    // read out text
                    while ((line = modelReader.readLine()) != null) {
                        if (line.length() > 0) {
                            // break on TEXT
                            if (line.matches(".*</TEXT>.*")) {
                                textfound = false;
                                break;
                            }
                        }
                    }
                    if (textfound) {
                        throw new Exception("Premature end of file (" + guide[i] + ") </TEXT> not found.");
                    }
                    // read out footer
                    while ((line = modelReader.readLine()) != null) {
                        if (!line.contains("<MAKEINSTANCE") && !line.matches(".*<.LINK.*")) {
                            footer += line + "\n";
                        }
                    }
                } finally {
                    if (modelReader != null) {
                        modelReader.close();
                    }
                }
                String mergedpipes = output_folder.getCanonicalPath() + File.separator + guide[i].getName() + ".normalized-tokens";
                BufferedWriter mergedpipesfile = new BufferedWriter(new FileWriter(mergedpipes));
                try {
                    String winner_entity4merging = null;
                    int winner_annotation4mergeing = -1;
                    int last_token2merge = -1;
                    for (int linen = 0; linen < tokenFileStringArr.get(0).length; linen++) {
                        if (winner_entity4merging == null) {
                            // get entity-system-pairs (in links, and weighted)
                            // and check matches (main attrib) for weighted voting
                            HashMap<String, Double> entity_voting = new HashMap<String, Double>();
                            for (int a = 0; a < annotations.size(); a++) {
                                String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                                HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                                if (pipesarr[2].startsWith("B-EVENT")) {
                                    String normid = event_map[a].get(attribs.get("eid"));
                                    if (merged_entities.contains(normid)) {
                                        continue;
                                    }
                                    if (weighted_entities[a].containsKey(normid)) {
                                        boolean hasVoted = false;
                                        for (String ev : entity_voting.keySet()) {
                                            if (ev.split("\\|")[2].equals(attribs.get("class"))) {
                                                entity_voting.put(ev, entity_voting.get(ev) + weighted_entities[a].get(normid));
                                                hasVoted = true;
                                                break;
                                            }
                                        }
                                        if (!hasVoted) {
                                            entity_voting.put(normid + "|" + a + "|" + attribs.get("class"), weighted_entities[a].get(normid));
                                        }
                                    }
                                }
                                if (pipesarr[2].startsWith("B-TIMEX3")) {
                                    String normid = timex_map[a].get(attribs.get("tid"));
                                    if (merged_entities.contains(normid)) {
                                        continue;
                                    }
                                    // in timex all the annotations of the same entity are added at the same time
                                    // put other timexes with the same id independently of where they start
                                    if (!entity_voting.keySet().toString().contains(normid + "|")) {
                                        for (int aw = 0; aw < annotations.size(); aw++) {
                                            if (weighted_entities[aw].containsKey(normid)) {
                                                boolean hasVoted = false;
                                                for (String ev : entity_voting.keySet()) {
                                                    if (ev.split("\\|")[2].equals(timex_extent_value_map[aw].get(normid).split("\\|")[2])) {
                                                        entity_voting.put(ev, entity_voting.get(ev) + weighted_entities[aw].get(normid));
                                                        hasVoted = true;
                                                        break;
                                                    }
                                                }
                                                if (!hasVoted) {
                                                    entity_voting.put(normid + "|" + aw + "|" + timex_extent_value_map[aw].get(normid).split("\\|")[2], weighted_entities[aw].get(normid));
                                                }
                                            }
                                        }

                                    }
                                }
                            }

                            if (entity_voting.size() > 0) {
                                winner_entity4merging = get_winner_entity(entity_voting);
                                // check for delay, and if it is the case choose alternative winner_entity4merging if doesn't overlap
                                if (winner_entity4merging.split("\\|")[0].startsWith("t")) {
                                    // vote extent and save winning entity to check delay...
                                    HashMap<String, Double> extent_voting = new HashMap<String, Double>();
                                    String winner_id = winner_entity4merging.split("\\|")[0];
                                    for (int aw = 0; aw < annotations.size(); aw++) {
                                        // annotation proposed the entity and proposed it WITH THE WINNING VALUE
                                        if (weighted_entities[aw].containsKey(winner_id) && timex_extent_value_map[aw].get(winner_id).split("\\|")[2].equals(winner_entity4merging.split("\\|")[2])) {
                                            boolean hasVoted = false;
                                            for (String exv : extent_voting.keySet()) {
                                                if (exv.split("\\|")[2].equals(timex_extent_value_map[aw].get(winner_id).split("\\|")[0] + "-" + timex_extent_value_map[aw].get(winner_id).split("\\|")[1])) {
                                                    extent_voting.put(exv, extent_voting.get(exv) + weighted_entities[aw].get(winner_id));
                                                    hasVoted = true;
                                                    break;
                                                }
                                            }
                                            if (!hasVoted) {
                                                extent_voting.put(winner_id + "|" + aw + "|" + timex_extent_value_map[aw].get(winner_id).split("\\|")[0] + "-" + timex_extent_value_map[aw].get(winner_id).split("\\|")[1], weighted_entities[aw].get(winner_id));
                                            }
                                        }
                                    }
                                    String winner_extent = get_winner_entity(extent_voting);
                                    //translate to winner_entity
                                    winner_entity4merging = winner_extent.split("\\|")[0] + "|" + winner_extent.split("\\|")[1] + "|" + winner_entity4merging.split("\\|")[2];

                                    // check delay with the winning extent
                                    if (Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[0]) > linen) {
                                        // too complicated just consider extra events...
                                        //int delay_until = Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[0]);
                                        entity_voting.remove(winner_entity4merging);
                                        winner_entity4merging = null;
                                        while (entity_voting.size() > 0) {
                                            winner_entity4merging = get_winner_entity(entity_voting);
                                            // too complicated... just avoid other timexes...
                                            if (winner_entity4merging.split("\\|")[0].startsWith("t")) { // || Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[1]) > delay_until) {
                                                entity_voting.remove(winner_entity4merging);
                                                winner_entity4merging = null;
                                            } else {
                                                break;
                                            }
                                        }
                                    }
                                }
                            }





                            if (winner_entity4merging == null) {
                                // entity needs >0.5 support (as entity), then events prefer 1 tok, timexes weighted vote extent
                                // entity tie (take first), voting extent
                                // DEPRECATED: This is not science... strategy (prefer longer, less than 5, the, preps), events prefer 1 tok
                                entity_voting = new HashMap<String, Double>();
                                for (int a = 0; a < annotations.size(); a++) {
                                    //System.out.println(tokenFileStringArr.get(a)[linen]);
                                    String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                                    HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                                    if (pipesarr[2].startsWith("B-EVENT")) {
                                        String normid = event_map[a].get(attribs.get("eid"));
                                        if (merged_entities.contains(normid)) {
                                            continue;
                                        }
                                        if (!weighted_entities[a].containsKey(normid)) {
                                            boolean hasVoted = false;
                                            for (String ev : entity_voting.keySet()) {
                                                if (ev.split("\\|")[2].equals(attribs.get("class"))) {
                                                    entity_voting.put(ev, entity_voting.get(ev) + weights[a]);
                                                    hasVoted = true;
                                                    break;
                                                }
                                            }
                                            if (!hasVoted) {
                                                entity_voting.put(normid + "|" + a + "|" + attribs.get("class"), weights[a]);
                                            }
                                        }
                                    }
                                    if (pipesarr[2].startsWith("B-TIMEX3")) {
                                        String normid = timex_map[a].get(attribs.get("tid"));
                                        if (merged_entities.contains(normid)) {
                                            continue;
                                        }
                                        // in timex all the annotations of the same entity are added at the same time
                                        // put other timexes with the same id independently of where they start
                                        if (!entity_voting.keySet().toString().contains(normid + "|")) {
                                            for (int aw = 0; aw < annotations.size(); aw++) {
                                                if (!weighted_entities[aw].containsKey(normid) && timex_extent_value_map[aw].containsKey(normid)) {
                                                    boolean hasVoted = false;
                                                    for (String ev : entity_voting.keySet()) {
                                                        if (ev.split("\\|")[2].equals(timex_extent_value_map[aw].get(normid).split("\\|")[2])) {
                                                            entity_voting.put(ev, entity_voting.get(ev) + weights[aw]);
                                                            hasVoted = true;
                                                            break;
                                                        }
                                                    }
                                                    if (!hasVoted) {
                                                        entity_voting.put(normid + "|" + aw + "|" + timex_extent_value_map[aw].get(normid).split("\\|")[2], weights[aw]);
                                                    }
                                                }
                                            }

                                        }
                                    }
                                }

                                if (entity_voting.size() > 0) {
                                    winner_entity4merging = get_winner_entity(entity_voting);
                                    //System.out.println("1a:"+winner_entity4merging +"\n1aaa:"+entity_voting.toString());

                                    // check for enough support
                                    if (entity_voting.get(winner_entity4merging) < min_et) {
                                        winner_entity4merging = null;
                                    } else {
                                        // check for delay, and if it is the case choose alternative winner_entity4merging if doesn't overlap
                                        if (winner_entity4merging.split("\\|")[0].startsWith("t")) {
                                            // vote extent and save winning entity to check delay...
                                            HashMap<String, Double> extent_voting = new HashMap<String, Double>();
                                            String winner_id = winner_entity4merging.split("\\|")[0];
                                            for (int aw = 0; aw < annotations.size(); aw++) {
                                                // annotation proposed the entity and proposed it WITH THE WINNING VALUE
                                                if (!weighted_entities[aw].containsKey(winner_id) && timex_extent_value_map[aw].containsKey(winner_id) && timex_extent_value_map[aw].get(winner_id).split("\\|")[2].equals(winner_entity4merging.split("\\|")[2])) {
                                                    boolean hasVoted = false;
                                                    for (String exv : extent_voting.keySet()) {
                                                        if (exv.split("\\|")[2].equals(timex_extent_value_map[aw].get(winner_id).split("\\|")[0] + "-" + timex_extent_value_map[aw].get(winner_id).split("\\|")[1])) {
                                                            extent_voting.put(exv, extent_voting.get(exv) + weights[aw]);
                                                            hasVoted = true;
                                                            break;
                                                        }
                                                    }
                                                    if (!hasVoted) {
                                                        extent_voting.put(winner_id + "|" + aw + "|" + timex_extent_value_map[aw].get(winner_id).split("\\|")[0] + "-" + timex_extent_value_map[aw].get(winner_id).split("\\|")[1], weights[aw]);
                                                    }
                                                }
                                            }

                                            String winner_extent = get_winner_entity(extent_voting);

                                            //translate to winner_entity
                                            winner_entity4merging = winner_extent.split("\\|")[0] + "|" + winner_extent.split("\\|")[1] + "|" + winner_entity4merging.split("\\|")[2];
                                            if (Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[0]) > linen) {
                                                // too complicated just consider extra events...
                                                //int delay_until = Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[0]);
                                                entity_voting.remove(winner_entity4merging);
                                                winner_entity4merging = null;
                                                while (entity_voting.size() > 0) {
                                                    winner_entity4merging = get_winner_entity(entity_voting);
                                                    // too complicated... just avoid other timexes...
                                                    if (winner_entity4merging.split("\\|")[0].startsWith("t")) { // || Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[1]) > delay_until) {
                                                        entity_voting.remove(winner_entity4merging);
                                                        winner_entity4merging = null;
                                                    } else {
                                                        break;
                                                    }
                                                }
                                                //recheck support
                                                if (entity_voting != null && winner_entity4merging != null && entity_voting.get(winner_entity4merging) < min_et) {
                                                    winner_entity4merging = null;
                                                }
                                            }
                                        }
                                    }


                                    // check possible overlapping with next weighted_entities
                                    // check if the winner has overlapping link-entities and if so discard and choose next winner (there is no next winner)
                                    // Since it needs more than 0.5 support
                                    if (winner_entity4merging != null) {
                                        if (winner_entity4merging.split("\\|")[0].startsWith("t")) {
                                            int winner_lasttoken = Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[1]);
                                            // search for weighted_entities in the timex span
                                            for (int tokn = linen + 1; tokn < winner_lasttoken; tokn++) {
                                                for (int a = 0; a < annotations.size(); a++) {
                                                    String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                                                    HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                                                    if (pipesarr[2].startsWith("B-EVENT")) {
                                                        String normid = event_map[a].get(attribs.get("eid"));
                                                        if (merged_entities.contains(normid)) {
                                                            continue;
                                                        }
                                                        if (weighted_entities[a].containsKey(normid)) {
                                                            winner_entity4merging = null;
                                                            break;
                                                        }
                                                    }
                                                    if (pipesarr[2].startsWith("B-TIMEX3")) {
                                                        String normid = timex_map[a].get(attribs.get("tid"));
                                                        if (merged_entities.contains(normid)) {
                                                            continue;
                                                        }
                                                        if (weighted_entities[a].containsKey(normid)) {
                                                            winner_entity4merging = null;
                                                            break;
                                                        }
                                                    }
                                                }
                                                if (winner_entity4merging == null) {
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                }
                            }







                            // write the final merging
                            if (winner_entity4merging == null) {
                                String[] pipesarr = tokenFileStringArr.get(0)[linen].split("\\|");
                                mergedpipesfile.write(pipesarr[0] + "|" + pipesarr[1] + "|O|-\n");
                                winner_entity4merging = null;
                            } else {
                                winner_annotation4mergeing = Integer.parseInt(winner_entity4merging.split("\\|")[1]);
                                String[] pipesarr = tokenFileStringArr.get(winner_annotation4mergeing)[linen].split("\\|");
                                HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                                // merge an event
                                if (winner_entity4merging.substring(0, 1).equals("e")) {
                                    HashMap<String, String> attribs2 = null;
                                    if (event_attr_from_mk_map[winner_annotation4mergeing].containsKey(attribs.get("eid"))) {
                                        attribs2 = XmlAttribs.parseAttrs(event_attr_from_mk_map[winner_annotation4mergeing].get(attribs.get("eid")));
                                    }
                                    merged_entities.add(event_map[winner_annotation4mergeing].get(attribs.get("eid")));
                                    mergedpipesfile.write(tokenFileStringArr.get(winner_annotation4mergeing)[linen].replaceFirst("eid=\"[^\"]*\"", "eid=\"" + event_map[winner_annotation4mergeing].get(attribs.get("eid")) + "\"") + "\n");
                                    ArrayList<String> mks = event_mk_index[winner_annotation4mergeing].get(event_map[winner_annotation4mergeing].get(attribs.get("eid")));
                                    if (mks != null) {
                                        for (int mki = 0; mki < mks.size(); mki++) {
                                            if (makeinstances.containsKey(mks.get(mki))) {
                                                String mkstring = "<MAKEINSTANCE eiid=\"" + mks.get(mki) + "\" ";
                                                mkstring += "eventID=\"" + event_map[winner_annotation4mergeing].get(attribs.get("eid")) + "\" ";
                                                if (attribs.get("pos") != null) {
                                                    mkstring += "pos=\"" + attribs.get("pos") + "\" ";
                                                } else {
                                                    if (attribs2.get("pos") != null) {
                                                        mkstring += "pos=\"" + attribs2.get("pos") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("tense") != null) {
                                                    mkstring += "tense=\"" + attribs.get("tense") + "\" ";
                                                } else {
                                                    if (attribs2.get("tense") != null) {
                                                        mkstring += "tense=\"" + attribs2.get("tense") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("aspect") != null) {
                                                    mkstring += "aspect=\"" + attribs.get("aspect") + "\" ";
                                                } else {
                                                    if (attribs2.get("aspect") != null) {
                                                        mkstring += "aspect=\"" + attribs2.get("aspect") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("polarity") != null) {
                                                    mkstring += "polarity=\"" + attribs.get("polarity") + "\" ";
                                                } else {
                                                    if (attribs2.get("polarity") != null) {
                                                        mkstring += "polarity=\"" + attribs2.get("polarity") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("modality") != null) {
                                                    mkstring += "modality=\"" + attribs.get("modality") + "\" ";
                                                } else {
                                                    if (attribs2.get("modality") != null) {
                                                        mkstring += "modality=\"" + attribs2.get("modality") + "\" ";
                                                    }
                                                }
                                                mkstring += "/>";
                                                makeinstances.put(mks.get(mki), mkstring); // update makeinstances with attribs
                                            }
                                        }
                                    }
                                    winner_entity4merging = null;
                                } else {
                                    //merge a timex
                                    merged_entities.add(timex_map[winner_annotation4mergeing].get(attribs.get("tid")));
                                    mergedpipesfile.write(tokenFileStringArr.get(winner_annotation4mergeing)[linen].replaceFirst("tid=\"[^\"]*\"", "tid=\"" + timex_map[winner_annotation4mergeing].get(attribs.get("tid")) + "\"") + "\n");
                                    // add merging, and control until which token are we merging
                                    last_token2merge = Integer.parseInt(timex_extent_value_map[winner_annotation4mergeing].get(timex_map[winner_annotation4mergeing].get(attribs.get("tid"))).split("\\|")[1]);
                                    if (last_token2merge == linen) {
                                        winner_entity4merging = null;
                                        last_token2merge = -1;
                                    }
                                }
                            }
                        } else {
                            mergedpipesfile.write(tokenFileStringArr.get(winner_annotation4mergeing)[linen] + "\n");
                            if (last_token2merge == linen) {
                                winner_entity4merging = null;
                                last_token2merge = -1;
                            }
                        }
                    }
                } finally {
                    if (mergedpipesfile != null) {
                        mergedpipesfile.close();
                    }
                }

                // of course add t0 (DCT)
                merged_entities.add("t0");
                create_merged_output(mergedpipes, makeinstances, links, merged_entities, header, footer, last_text_blanks, output_folder + File.separator + guide[i].getName());
                if (System.getProperty("DEBUG") == null || System.getProperty("DEBUG").equalsIgnoreCase("false")) {
                    (new File(mergedpipes)).delete();
                }
            }
            //delete data
            for (int i = 0; i < guide.length; i++) {
                for (int a = 0; a < annotations.size(); a++) {
                    File annot = annotations.get(a)[i];
                    File ftdir = new File(annot.getCanonicalPath() + "-data");
                    if (ftdir.exists()) {
                        FileUtils.deleteRecursively(ftdir);
                    }
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TimeML_Merger):\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
        }
    }

    public static void merge_bottom_up(ArrayList<File[]> annotations, double[] weights, File output_folder, double min_et, String link_threshold, String links_merge_option) {
        try {
            // use annotations[0] as key-guide
            File[] guide = annotations.get(0);
            // for each file
            for (int i = 0; i < guide.length; i++) {
                System.out.println("Merging bottom up..." + guide[i].getName());
                HashMap<String, String>[] timex_map = new HashMap[annotations.size()]; // original-normalized id
                HashMap<String, String>[] event_map = new HashMap[annotations.size()]; // original-normalized id
                HashMap<String, String>[] mk_map = new HashMap[annotations.size()]; // original-normalized id
                HashMap<String, String>[] event_attr_from_mk_map = new HashMap[annotations.size()]; // original-id-mk attributes
                HashMap<String, String>[] timex_extent_value_map = new HashMap[annotations.size()]; // normalized-id-extent-value-map
                HashMap<String, String>[] event_class_map = new HashMap[annotations.size()]; // normalized-id-extent-value-map
                HashMap<String, ArrayList<String>>[] event_mk_index = new HashMap[annotations.size()]; // normalied id - related mks
                HashMap<String, String> normmk_normevent_map = new HashMap<String, String>();
                HashMap<String, Double> weighted_links = new HashMap<String, Double>();
                HashMap<String, String> links_support_map = new HashMap<String, String>();
                HashMap<String, String> links_support_map_names = new HashMap<String, String>();
                HashSet<String> merged_entities = new HashSet<String>();
                ArrayList<String> xmlFileString = new ArrayList();
                ArrayList<String[]> tokenFileStringArr = new ArrayList();
                int last_eid = 1;
                int open_event = 0;
                int last_tid = 1;
                int[] open_timex = new int[annotations.size()];
                int[] last_tid_local = new int[annotations.size()];
                String last_text_blanks = "";
                // for each annotation
                for (int a = 0; a < annotations.size(); a++) {
                    File annot = annotations.get(a)[i];
                    File ftdir = new File(annot.getCanonicalPath() + "-data");
                    if (ftdir.exists()) {
                        FileUtils.deleteRecursively(ftdir);
                    }
                    if (!ftdir.mkdirs()) {  // mkdirs creates many parent dirs if needed
                        throw new Exception("Directory not created...");
                    }
                    File workingfile = new File(ftdir + File.separator + annot.getName());
                    FileUtils.copyFileUtil(annot, workingfile);
                    XMLFile xmlfile = new XMLFile();
                    xmlfile.loadFile(workingfile);
                    String plainfile = xmlfile.toPlain(); // only <text>
                    String output = Tokenizer.run(plainfile);
                    output = addSpacing(output, plainfile);
                    output = merge_tok_n_tml(output, workingfile.getCanonicalPath());
                    last_text_blanks = get_last_text_blanks(plainfile);
                    //System.out.println("last blanks:"+last_text_blanks+".");
                    tokenFileStringArr.add(FileUtils.readFileAsString(output, "UTF-8").split("\\n"));
                    xmlFileString.add(FileUtils.readFileAsString(workingfile.getCanonicalPath(), "UTF-8"));
                    event_map[a] = new HashMap<String, String>();
                    timex_map[a] = new HashMap<String, String>();
                    mk_map[a] = new HashMap<String, String>();
                    event_attr_from_mk_map[a] = new HashMap<String, String>();
                    timex_extent_value_map[a] = new HashMap<String, String>();
                    event_class_map[a] = new HashMap<String, String>();
                    event_mk_index[a] = new HashMap<String, ArrayList<String>>();
                    open_timex[a] = 0;
                    last_tid_local[a] = 0;
                }
                for (int linen = 0; linen < tokenFileStringArr.get(0).length; linen++) {
                    open_event = 0; // restart in every iteration (multi-token are not considered)
                    String last_token = null; // check all files are equal
                    for (int a = 0; a < annotations.size(); a++) {
                        String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                        if (last_token == null) {
                            last_token = pipesarr[0];
                        } else {
                            if (pipesarr[0].equals(last_token)) {
                                last_token = pipesarr[0];
                            } else {
                                throw new Exception("Tokens doesn't match at line: " + linen + " (" + last_token + ")");
                            }
                        }
                        HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                        if (pipesarr[2].matches("B-EVENT")) {
                            if (open_event == 0) {
                                open_event = last_eid;
                                last_eid++;
                            }
                            event_map[a].put(attribs.get("eid"), "e" + open_event);
                            event_class_map[a].put("e" + open_event, attribs.get("class"));
                        }
                        // for timex maintain open_timex[] first
                        if (pipesarr[2].equals("O") || pipesarr[2].equals("B-TIMEX3")) {
                            open_timex[a] = 0; // restart
                        }
                    }
                    for (int a = 0; a < annotations.size(); a++) {
                        String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                        HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                        if (pipesarr[2].equals("B-TIMEX3")) {
                            open_timex[a] = last_tid; // new by default
                            // try to match with still open timexes
                            for (int at = 0; at < annotations.size(); at++) {
                                // open and not used in other annotations
                                if (a != at && open_timex[at] != 0 && last_tid_local[a] < open_timex[at]) {
                                    open_timex[a] = open_timex[at];
                                    break;
                                }
                            }
                            last_tid_local[a] = open_timex[a];
                            //new (not paired)
                            if (open_timex[a] == last_tid) {
                                last_tid++;
                            }
                            timex_map[a].put(attribs.get("tid"), "t" + open_timex[a]);
                            timex_extent_value_map[a].put("t" + open_timex[a], linen + "|" + linen + "|" + attribs.get("value"));
                        }
                        // update extent of if open timex
                        if (pipesarr[2].matches("I-TIMEX3")) {
                            timex_extent_value_map[a].put("t" + open_timex[a], timex_extent_value_map[a].get("t" + open_timex[a]).replaceFirst("\\|[0-9]+\\|", "|" + linen + "|"));
                        }
                    }
                }
                // normalize entities given a map
                for (int a = 0; a < annotations.size(); a++) {
                    String annotname = annotations.get(a)[i].getCanonicalPath().substring(0, annotations.get(a)[i].getCanonicalPath().lastIndexOf(File.separator));
                    annotname = annotname.substring(annotname.lastIndexOf(File.separator) + 1);
                    String tmp = xmlFileString.get(a);
                    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                    DocumentBuilder db = dbf.newDocumentBuilder();
                    Document doc = db.parse(annotations.get(a)[i]);
                    doc.getDocumentElement().normalize();
                    String dctid = null;
                    Element dct = ((Element) ((NodeList) ((Element) doc.getElementsByTagName("DCT").item(0)).getElementsByTagName("TIMEX3")).item(0));
                    if (dct != null) {
                        dctid = dct.getAttribute("tid");
                        timex_map[a].put(dctid, "t0");
                        // normalize to t0
                        tmp = tmp.replaceAll("(<DCT><TIMEX3[^>]*tid=\")" + dct.getAttribute("tid") + "(\"[^>]*>)", "$1t0$2");
                    }
                    NodeList text = doc.getElementsByTagName("TEXT");
                    if (text.getLength() > 1) {
                        throw new Exception("More than one TEXT tag found.");
                    }
                    Element TextElmnt = (Element) text.item(0); // If not ELEMENT NODE will throw exception
                    // normalize events (entity by entity in order) ---- // deprecated solution: e1=e3 -- e3=e1 map problem. //for(String e:event_map[a].keySet()){tmp=tmp.replaceAll("(eid|eventID)=\""+e+"\"", event_map[a].get(e));}
                    NodeList current_node = TextElmnt.getElementsByTagName("EVENT");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        tmp = tmp.replaceAll("(<EVENT[^>]*eid=\")" + element.getAttribute("eid") + "(\"[^>]*>)", "$1n" + event_map[a].get(element.getAttribute("eid")) + "$2");
                    }
                    tmp = tmp.replaceAll("(<EVENT[^>]*eid=\")n", "$1");
                    // normalize timexes
                    current_node = TextElmnt.getElementsByTagName("TIMEX3");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        tmp = tmp.replaceAll("(<TIMEX3[^>]*tid=\")" + element.getAttribute("tid") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("tid")) + "$2");
                        if (element.hasAttribute("anchorTimeID")) {
                            tmp = tmp.replaceAll("(<TIMEX3[^>]*anchorTimeID=\")" + element.getAttribute("anchorTimeID") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("anchorTimeID")) + "$2");
                        }
                        if (element.hasAttribute("beginPoint")) {
                            tmp = tmp.replaceAll("(<TIMEX3[^>]*beginPoint=\")" + element.getAttribute("beginPoint") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("beginPoint")) + "$2");
                        }
                        if (element.hasAttribute("endPoint")) {
                            tmp = tmp.replaceAll("(<TIMEX3[^>]*endPoint=\")" + element.getAttribute("endPoint") + "(\"[^>]*>)", "$1n" + timex_map[a].get(element.getAttribute("endPoint")) + "$2");
                        }
                    }
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*tid=\")n", "$1");
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*anchorTimeID=\")n", "$1");
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*beginPoint=\")n", "$1");
                    tmp = tmp.replaceAll("(<TIMEX3[^>]*endPoint=\")n", "$1");
                    // normalize and map makeinstances
                    HashSet<String> mks_new_ids = new HashSet<String>();
                    current_node = null;
                    current_node = doc.getElementsByTagName("MAKEINSTANCE");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        String mapped_event = event_map[a].get(element.getAttribute("eventID"));
                        if (!event_mk_index[a].containsKey(mapped_event)) {
                            // add attribs for the first
                            String event_attr_from_mk = "";
                            if (element.hasAttribute("pos")) {
                                event_attr_from_mk += "pos=\"" + element.getAttribute("pos") + "\" ";
                            }
                            if (element.hasAttribute("tense")) {
                                event_attr_from_mk += "tense=\"" + element.getAttribute("tense") + "\" ";
                            }
                            if (element.hasAttribute("aspect")) {
                                event_attr_from_mk += "aspect=\"" + element.getAttribute("aspect") + "\" ";
                            }
                            if (element.hasAttribute("polarity")) {
                                event_attr_from_mk += "polarity=\"" + element.getAttribute("polarity") + "\" ";
                            }
                            if (element.hasAttribute("modality")) {
                                event_attr_from_mk += "modality=\"" + element.getAttribute("modality") + "\" ";
                            }
                            event_attr_from_mk_map[a].put(element.getAttribute("eventID"), event_attr_from_mk);
                            ArrayList<String> mks = new ArrayList<String>();
                            if (mks_new_ids.contains(mapped_event.replaceFirst("e", "ei"))) {
                                throw new Exception("Duplicated eiid in mks: " + mapped_event.replaceFirst("e", "ei") + "  " + annotations.get(a)[i]);
                            }
                            normmk_normevent_map.put(mapped_event.replaceFirst("e", "ei"), mapped_event);
                            mks_new_ids.add(mapped_event.replaceFirst("e", "ei"));
                            mks.add(mapped_event.replaceFirst("e", "ei"));
                            event_mk_index[a].put(mapped_event, mks);
                            mk_map[a].put(element.getAttribute("eiid"), mapped_event.replaceFirst("e", "ei"));
                            tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*>)", "$1n" + mapped_event + "$2");
                            tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*>)", "$1n" + mapped_event + "$2");
                            tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\")" + element.getAttribute("eiid") + "(\"[^>]*>)", "$1n" + mapped_event.replaceFirst("e", "ei") + "$2");
                        } else {
                            if (mks_new_ids.contains(mapped_event.replaceFirst("e", "ei"))) {
                                int num = 1000000 + Integer.parseInt(mapped_event.substring(1));
                                while (mks_new_ids.contains("ei" + num)) {
                                    num++;
                                }
                                mks_new_ids.add("ei" + num);
                                normmk_normevent_map.put("ei" + num, mapped_event);
                                event_mk_index[a].get(mapped_event).add("ei" + num);
                                mk_map[a].put(element.getAttribute("eiid"), "ei" + num);
                                tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*>)", "$1n" + mapped_event + "$2");
                                tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*[^>]*eventID=\")" + element.getAttribute("eventID") + "(\"[^>]*eiid=\"" + element.getAttribute("eiid") + "\"[^>]*>)", "$1n" + mapped_event + "$2");
                                tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\")" + element.getAttribute("eiid") + "(\"[^>]*>)", "$1n" + "ei" + num + "$2");
                            } else {
                                throw new Exception("Extrange... mkinstance");
                            }
                        }
                    }
                    tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eiid=\")n", "$1");
                    tmp = tmp.replaceAll("(<MAKEINSTANCE[^>]*eventID=\")n", "$1");
                    // normalize links
                    current_node = doc.getElementsByTagName("TLINK");
                    for (int s = 0; s < current_node.getLength(); s++) {
                        Element element = (Element) current_node.item(s);
                        String relType = element.getAttribute("relType");
                        if (relType.matches("(DURING|DURING_INV|IDENTITY)")) {
                            relType = "SIMULTANEOUS";
                        }
                        String entity1 = null;
                        String entity2 = null;
                        // event-event
                        if (element.hasAttribute("eventInstanceID") && element.hasAttribute("relatedToEventInstance")) {
                            if (!mk_map[a].containsKey(element.getAttribute("eventInstanceID"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("eventInstanceID"));
                            }
                            if (!mk_map[a].containsKey(element.getAttribute("relatedToEventInstance"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("relatedToEventInstance"));
                            }
                            entity1 = mk_map[a].get(element.getAttribute("eventInstanceID"));
                            entity2 = mk_map[a].get(element.getAttribute("relatedToEventInstance"));
                            // Order by id (for normalization)
                            if (Integer.parseInt(entity1.substring(2)) > Integer.parseInt(entity2.substring(2))) {
                                entity1 = entity2;
                                entity2 = mk_map[a].get(element.getAttribute("eventInstanceID"));
                                relType = reverseRelationCategory(relType);
                            }
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*eventInstanceID=\")" + element.getAttribute("eventInstanceID") + "(\"[^>]*>)", "$1" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relatedToEventInstance=\")" + element.getAttribute("relatedToEventInstance") + "(\"[^>]*>)", "$1" + entity2 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relType=\")" + element.getAttribute("relType") + "(\"[^>]*>)", "$1" + relType + "$2");
                        }
                        // event-time
                        if (element.hasAttribute("eventInstanceID") && element.hasAttribute("relatedToTime")) {
                            if (!mk_map[a].containsKey(element.getAttribute("eventInstanceID"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("eventInstanceID"));
                            }
                            if (!timex_map[a].containsKey(element.getAttribute("relatedToTime"))) {
                                throw new Exception("Timex not found it file: " + element.getAttribute("relatedToTime"));
                            }
                            entity1 = mk_map[a].get(element.getAttribute("eventInstanceID"));
                            entity2 = timex_map[a].get(element.getAttribute("relatedToTime"));
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*eventInstanceID=\")" + element.getAttribute("eventInstanceID") + "(\"[^>]*>)", "$1" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relatedToTime=\")" + element.getAttribute("relatedToTime") + "(\"[^>]*>)", "$1" + entity2 + "$2");
                        }
                        if (element.hasAttribute("timeID") && element.hasAttribute("relatedToEventInstance")) {
                            if (!mk_map[a].containsKey(element.getAttribute("relatedToEventInstance"))) {
                                throw new Exception("Event instance not found it file: " + element.getAttribute("relatedToEventInstance"));
                            }
                            if (!timex_map[a].containsKey(element.getAttribute("timeID"))) {
                                throw new Exception("Timex not found it file: " + element.getAttribute("timeID"));
                            }
                            entity1 = mk_map[a].get(element.getAttribute("relatedToEventInstance"));
                            entity2 = timex_map[a].get(element.getAttribute("timeID"));
                            relType = reverseRelationCategory(relType);
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*)timeID=\"" + element.getAttribute("timeID") + "(\"[^>]*>)", "$1eventInstanceID=\"" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*)relatedToEventInstance=\"" + element.getAttribute("relatedToEventInstance") + "(\"[^>]*>)", "$1relatedToTime=\"" + entity2 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relType=\")" + element.getAttribute("relType") + "(\"[^>]*>)", "$1" + relType + "$2");
                        }
                        // time-time
                        if (element.hasAttribute("timeID") && element.hasAttribute("relatedToTime")) {
                            if (!timex_map[a].containsKey(element.getAttribute("relatedToTime"))) {
                                throw new Exception("Timex instance not found it file: " + element.getAttribute("relatedToTime"));
                            }
                            if (!timex_map[a].containsKey(element.getAttribute("timeID"))) {
                                throw new Exception("Timex not found it file: " + element.getAttribute("timeID"));
                            }
                            entity1 = timex_map[a].get(element.getAttribute("timeID"));
                            entity2 = timex_map[a].get(element.getAttribute("relatedToTime"));
                            // Order by id (for normalization)
                            if (Integer.parseInt(entity1.substring(1)) > Integer.parseInt(entity2.substring(1))) {
                                entity1 = entity2;
                                entity2 = timex_map[a].get(element.getAttribute("timeID"));
                                relType = reverseRelationCategory(relType);
                            }
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*timeID=\")" + element.getAttribute("timeID") + "(\"[^>]*>)", "$1" + entity1 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relatedToTime=\")" + element.getAttribute("relatedToTime") + "(\"[^>]*>)", "$1" + entity2 + "$2");
                            tmp = tmp.replaceAll("(<TLINK[^>]*lid=\"" + element.getAttribute("lid") + "\"[^>]*[^>]*relType=\")" + element.getAttribute("relType") + "(\"[^>]*>)", "$1" + relType + "$2");
                        }
                        // save and calculate link-weight (#triples*100+#doubles)
                        // with weights #triples*(100+weriht_i/10+#doubles*(1+weriht_i/10))
                        // for ties sort higher weighted annotations first (weight is normalized to sum 100)
                        // save and calculate link-weight (#triples*(n*10)+#doubles*n)
                        // with weights #triples*(n*10)+#doubles*n+(system weights)
                        // for ties sort higher weighted annotations first (weight is normalized to sum 100)
                        // upadte links support map
                        if (links_support_map.containsKey(entity1 + "|" + entity2 + "|" + relType)) {
                            links_support_map.put(entity1 + "|" + entity2 + "|" + relType, links_support_map.get(entity1 + "|" + entity2 + "|" + relType) + ";" + a);
                            links_support_map_names.put(entity1 + "|" + entity2 + "|" + relType, links_support_map_names.get(entity1 + "|" + entity2 + "|" + relType) + ";" + annotname);
                        } else {
                            // add new triplet
                            links_support_map.put(entity1 + "|" + entity2 + "|" + relType, "" + a);
                            links_support_map_names.put(entity1 + "|" + entity2 + "|" + relType, "" + annotname);
                        }
                        // check triplet
                        if (weighted_links.containsKey(entity1 + "|" + entity2 + "|" + relType)) {
                            weighted_links.put(entity1 + "|" + entity2 + "|" + relType, weighted_links.get(entity1 + "|" + entity2 + "|" + relType) + (annotations.size()+1) * weights[a]);
                        } else {
                            // add new triplet
                            weighted_links.put(entity1 + "|" + entity2 + "|" + relType, (annotations.size()+1) * weights[a]);
                        }
                        //check matching doublet
                        String doublets = weighted_links.keySet().toString();
                        Pattern pattern = Pattern.compile("" + entity1 + "\\|" + entity2 + "\\|[^,\\]]+");
                        Matcher matcher = pattern.matcher(doublets);
                        if (matcher.find()) {
                            for (int dblt = 0; dblt <= matcher.groupCount(); dblt++) {
                                String matchdbl = matcher.group(dblt);
                                weighted_links.put(matchdbl, weighted_links.get(matchdbl) + weights[a]);
                            }
                        }
                    }
                    // For safety ALINKS and SLINKs are removed
                    tmp = tmp.replaceAll("<[AS]LINK.*", "");
                    // save normalized annotation
                    FileUtils.writeFileFromString(tmp, annotations.get(a)[i].getCanonicalPath() + "-data" + File.separator + annotations.get(a)[i].getName() + ".normalized");
                }
                String line;
                boolean textfound = false;
                String header = "";
                String footer = "";
                BufferedReader modelReader = new BufferedReader(new FileReader(guide[i]));
                try {
                    // read out header
                    while ((line = modelReader.readLine()) != null) {
                        if (line.length() > 0) {
                            if (line.matches(".*<TEXT>.*")) {
                                textfound = true;
                                break;
                            }
                        }
                        header += line + "\n";
                    }
                    if (!textfound) {
                        throw new Exception("Premature end of file (" + guide[i] + ") <TEXT> not found.");
                    }
                    // read out text
                    while ((line = modelReader.readLine()) != null) {
                        if (line.length() > 0) {
                            // break on TEXT
                            if (line.matches(".*</TEXT>.*")) {
                                textfound = false;
                                break;
                            }
                        }
                    }
                    if (textfound) {
                        throw new Exception("Premature end of file (" + guide[i] + ") </TEXT> not found.");
                    }
                    // read out footer
                    while ((line = modelReader.readLine()) != null) {
                        if (!line.contains("<MAKEINSTANCE") && !line.matches(".*<.LINK.*")) {
                            footer += line + "\n";
                        }
                    }
                } finally {
                    if (modelReader != null) {
                        modelReader.close();
                    }
                }

                String mergedpipes = output_folder.getCanonicalPath() + File.separator + guide[i].getName() + ".normalized-tokens";
                BufferedWriter mergedpipesfile = new BufferedWriter(new FileWriter(mergedpipes));
                HashMap<String, String> makeinstances = new HashMap<String, String>();
                try {
                    String winner_entity4merging = null;
                    int winner_annotation4mergeing = -1;
                    int last_token2merge = -1;
                    for (int linen = 0; linen < tokenFileStringArr.get(0).length; linen++) {
                        if (winner_entity4merging == null) {
                            HashMap<String, Double> entity_voting = new HashMap<String, Double>();

                            // entity needs >et_s support (as entity), then events prefer 1 tok, timexes weighted vote extent
                            // entity tie (take first), voting extent
                            // DEPRECATED: This is not science... strategy (prefer longer, less than 5, the, preps), events prefer 1 tok
                            entity_voting = new HashMap<String, Double>();
                            for (int a = 0; a < annotations.size(); a++) {
                                //System.out.println(tokenFileStringArr.get(a)[linen]);
                                String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                                HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                                if (pipesarr[2].startsWith("B-EVENT")) {
                                    String normid = event_map[a].get(attribs.get("eid"));
                                    if (merged_entities.contains(normid)) {
                                        continue;
                                    }
                                    boolean hasVoted = false;
                                    for (String ev : entity_voting.keySet()) {
                                        if (ev.split("\\|")[2].equals(attribs.get("class"))) {
                                            entity_voting.put(ev, entity_voting.get(ev) + weights[a]);
                                            hasVoted = true;
                                            break;
                                        }
                                    }
                                    if (!hasVoted) {
                                        entity_voting.put(normid + "|" + a + "|" + attribs.get("class"), weights[a]);
                                    }
                                }
                                if (pipesarr[2].startsWith("B-TIMEX3")) {
                                    String normid = timex_map[a].get(attribs.get("tid"));
                                    if (merged_entities.contains(normid)) {
                                        continue;
                                    }
                                    // in timex all the annotations of the same entity are added at the same time
                                    // put other timexes with the same id independently of where they start
                                    if (!entity_voting.keySet().toString().contains(normid + "|")) {
                                        for (int aw = 0; aw < annotations.size(); aw++) {
                                            if (timex_extent_value_map[aw].containsKey(normid)) {
                                                boolean hasVoted = false;
                                                for (String ev : entity_voting.keySet()) {
                                                    if (ev.split("\\|")[2].equals(timex_extent_value_map[aw].get(normid).split("\\|")[2])) {
                                                        entity_voting.put(ev, entity_voting.get(ev) + weights[aw]);
                                                        hasVoted = true;
                                                        break;
                                                    }
                                                }
                                                if (!hasVoted) {
                                                    entity_voting.put(normid + "|" + aw + "|" + timex_extent_value_map[aw].get(normid).split("\\|")[2], weights[aw]);
                                                }
                                            }
                                        }

                                    }
                                }
                            }

                            if (entity_voting.size() > 0) {
                                winner_entity4merging = get_winner_entity(entity_voting);
                                //System.out.println("1a:"+winner_entity4merging +"\n1aaa:"+entity_voting.toString());

                                // check for enough support
                                if (entity_voting.get(winner_entity4merging) < min_et) {
                                    winner_entity4merging = null;
                                } else {
                                    // check for delay, and if it is the case choose alternative winner_entity4merging if doesn't overlap
                                    if (winner_entity4merging.split("\\|")[0].startsWith("t")) {
                                        // vote extent and save winning entity to check delay...
                                        HashMap<String, Double> extent_voting = new HashMap<String, Double>();
                                        String winner_id = winner_entity4merging.split("\\|")[0];
                                        for (int aw = 0; aw < annotations.size(); aw++) {
                                            // annotation proposed the entity and proposed it WITH THE WINNING VALUE
                                            if (timex_extent_value_map[aw].containsKey(winner_id) && timex_extent_value_map[aw].get(winner_id).split("\\|")[2].equals(winner_entity4merging.split("\\|")[2])) {
                                                boolean hasVoted = false;
                                                for (String exv : extent_voting.keySet()) {
                                                    if (exv.split("\\|")[2].equals(timex_extent_value_map[aw].get(winner_id).split("\\|")[0] + "-" + timex_extent_value_map[aw].get(winner_id).split("\\|")[1])) {
                                                        extent_voting.put(exv, extent_voting.get(exv) + weights[aw]);
                                                        hasVoted = true;
                                                        break;
                                                    }
                                                }
                                                if (!hasVoted) {
                                                    extent_voting.put(winner_id + "|" + aw + "|" + timex_extent_value_map[aw].get(winner_id).split("\\|")[0] + "-" + timex_extent_value_map[aw].get(winner_id).split("\\|")[1], weights[aw]);
                                                }
                                            }
                                        }

                                        String winner_extent = get_winner_entity(extent_voting);

                                        //translate to winner_entity
                                        winner_entity4merging = winner_extent.split("\\|")[0] + "|" + winner_extent.split("\\|")[1] + "|" + winner_entity4merging.split("\\|")[2];
                                        if (Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[0]) > linen) {
                                            // too complicated just consider extra events...
                                            //int delay_until = Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[0]);
                                            entity_voting.remove(winner_entity4merging);
                                            winner_entity4merging = null;
                                            while (entity_voting.size() > 0) {
                                                winner_entity4merging = get_winner_entity(entity_voting);
                                                // too complicated... just avoid other timexes...
                                                if (winner_entity4merging.split("\\|")[0].startsWith("t")) { // || Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[1]) > delay_until) {
                                                    entity_voting.remove(winner_entity4merging);
                                                    winner_entity4merging = null;
                                                } else {
                                                    break;
                                                }
                                            }
                                            //recheck support
                                            if (entity_voting != null && winner_entity4merging != null && entity_voting.get(winner_entity4merging) < min_et) {
                                                winner_entity4merging = null;
                                            }
                                        }
                                    }
                                }


                                // check possible overlapping with next weighted_entities
                                // check if the winner has overlapping link-entities and if so discard and choose next winner (there is no next winner)
                                // Since it needs more than 0.5 support
                                if (winner_entity4merging != null) {
                                    if (winner_entity4merging.split("\\|")[0].startsWith("t")) {
                                        int winner_lasttoken = Integer.parseInt(timex_extent_value_map[Integer.parseInt(winner_entity4merging.split("\\|")[1])].get(winner_entity4merging.split("\\|")[0]).split("\\|")[1]);
                                        // search for weighted_entities in the timex span
                                        for (int tokn = linen + 1; tokn < winner_lasttoken; tokn++) {
                                            for (int a = 0; a < annotations.size(); a++) {
                                                String[] pipesarr = tokenFileStringArr.get(a)[linen].split("\\|");
                                                HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                                                if (pipesarr[2].startsWith("B-EVENT")) {
                                                    String normid = event_map[a].get(attribs.get("eid"));
                                                    if (merged_entities.contains(normid)) {
                                                        continue;
                                                    }
                                                }
                                                if (pipesarr[2].startsWith("B-TIMEX3")) {
                                                    String normid = timex_map[a].get(attribs.get("tid"));
                                                    if (merged_entities.contains(normid)) {
                                                        continue;
                                                    }
                                                }
                                            }
                                            if (winner_entity4merging == null) {
                                                break;
                                            }
                                        }
                                    }
                                }
                            }








                            // write the final merging
                            if (winner_entity4merging == null) {
                                String[] pipesarr = tokenFileStringArr.get(0)[linen].split("\\|");
                                mergedpipesfile.write(pipesarr[0] + "|" + pipesarr[1] + "|O|-\n");
                                winner_entity4merging = null;
                            } else {
                                winner_annotation4mergeing = Integer.parseInt(winner_entity4merging.split("\\|")[1]);
                                String[] pipesarr = tokenFileStringArr.get(winner_annotation4mergeing)[linen].split("\\|");
                                HashMap<String, String> attribs = XmlAttribs.parseAttrs(pipesarr[3]);
                                // merge an event
                                if (winner_entity4merging.substring(0, 1).equals("e")) {
                                    HashMap<String, String> attribs2 = null;
                                    if (event_attr_from_mk_map[winner_annotation4mergeing].containsKey(attribs.get("eid"))) {
                                        attribs2 = XmlAttribs.parseAttrs(event_attr_from_mk_map[winner_annotation4mergeing].get(attribs.get("eid")));
                                    }
                                    merged_entities.add(event_map[winner_annotation4mergeing].get(attribs.get("eid")));
                                    mergedpipesfile.write(tokenFileStringArr.get(winner_annotation4mergeing)[linen].replaceFirst("eid=\"[^\"]*\"", "eid=\"" + event_map[winner_annotation4mergeing].get(attribs.get("eid")) + "\"") + "\n");
                                    ArrayList<String> mks = event_mk_index[winner_annotation4mergeing].get(event_map[winner_annotation4mergeing].get(attribs.get("eid")));
                                    if (mks != null) {
                                        for (int mki = 0; mki < mks.size(); mki++) {
                                            if (!makeinstances.containsKey(mks.get(mki))) {
                                                makeinstances.put(mks.get(mki), "<MAKEINSTANCE eiid=\"" + mks.get(mki) + "\" />");
                                            }
                                            if (makeinstances.containsKey(mks.get(mki))) {
                                                String mkstring = "<MAKEINSTANCE eiid=\"" + mks.get(mki) + "\" ";
                                                mkstring += "eventID=\"" + event_map[winner_annotation4mergeing].get(attribs.get("eid")) + "\" ";
                                                if (attribs.get("pos") != null) {
                                                    mkstring += "pos=\"" + attribs.get("pos") + "\" ";
                                                } else {
                                                    if (attribs2.get("pos") != null) {
                                                        mkstring += "pos=\"" + attribs2.get("pos") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("tense") != null) {
                                                    mkstring += "tense=\"" + attribs.get("tense") + "\" ";
                                                } else {
                                                    if (attribs2.get("tense") != null) {
                                                        mkstring += "tense=\"" + attribs2.get("tense") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("aspect") != null) {
                                                    mkstring += "aspect=\"" + attribs.get("aspect") + "\" ";
                                                } else {
                                                    if (attribs2.get("aspect") != null) {
                                                        mkstring += "aspect=\"" + attribs2.get("aspect") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("polarity") != null) {
                                                    mkstring += "polarity=\"" + attribs.get("polarity") + "\" ";
                                                } else {
                                                    if (attribs2.get("polarity") != null) {
                                                        mkstring += "polarity=\"" + attribs2.get("polarity") + "\" ";
                                                    }
                                                }
                                                if (attribs.get("modality") != null) {
                                                    mkstring += "modality=\"" + attribs.get("modality") + "\" ";
                                                } else {
                                                    if (attribs2.get("modality") != null) {
                                                        mkstring += "modality=\"" + attribs2.get("modality") + "\" ";
                                                    }
                                                }
                                                mkstring += "/>";
                                                makeinstances.put(mks.get(mki), mkstring); // update makeinstances with attribs
                                            }
                                        }
                                    }
                                    winner_entity4merging = null;
                                } else {
                                    //merge a timex
                                    merged_entities.add(timex_map[winner_annotation4mergeing].get(attribs.get("tid")));
                                    mergedpipesfile.write(tokenFileStringArr.get(winner_annotation4mergeing)[linen].replaceFirst("tid=\"[^\"]*\"", "tid=\"" + timex_map[winner_annotation4mergeing].get(attribs.get("tid")) + "\"") + "\n");
                                    // add merging, and control until which token are we merging
                                    last_token2merge = Integer.parseInt(timex_extent_value_map[winner_annotation4mergeing].get(timex_map[winner_annotation4mergeing].get(attribs.get("tid"))).split("\\|")[1]);
                                    if (last_token2merge == linen) {
                                        winner_entity4merging = null;
                                        last_token2merge = -1;
                                    }
                                }
                            }
                        } else {
                            mergedpipesfile.write(tokenFileStringArr.get(winner_annotation4mergeing)[linen] + "\n");
                            if (last_token2merge == linen) {
                                winner_entity4merging = null;
                                last_token2merge = -1;
                            }
                        }
                    }
                } finally {
                    if (mergedpipesfile != null) {
                        mergedpipesfile.close();
                    }
                }


                // weighted links //System.out.println(weighted_links);
                String weightedlinks = saveWeightedLinks(weighted_links, links_support_map_names, annotations.get(0)[i], output_folder);
                String mergedlinks = LinkMerger.run(weightedlinks, links_merge_option);
                HashMap<String, String> links = new HashMap<String, String>();

                // of course add t0 (DCT)
                merged_entities.add("t0");

                obtainInfoFromMergedLinks_bottom_up(mergedlinks, weights, normmk_normevent_map, makeinstances, links, link_threshold, merged_entities);
                if (System.getProperty("DEBUG") == null || System.getProperty("DEBUG").equalsIgnoreCase("false")) {
                    (new File(weightedlinks)).delete();
                    (new File(mergedlinks)).delete();
                }

                create_merged_output(mergedpipes, makeinstances, links, merged_entities, header, footer, last_text_blanks, output_folder + File.separator + guide[i].getName());
                if (System.getProperty("DEBUG") == null || System.getProperty("DEBUG").equalsIgnoreCase("false")) {
                    (new File(mergedpipes)).delete();
                }
            }
            //delete data
            for (int i = 0; i < guide.length; i++) {
                for (int a = 0; a < annotations.size(); a++) {
                    File annot = annotations.get(a)[i];
                    File ftdir = new File(annot.getCanonicalPath() + "-data");
                    if (ftdir.exists()) {
                        FileUtils.deleteRecursively(ftdir);
                    }
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TimeML_Merger):\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
        }
    }

    public static String get_winner_entity(HashMap<String, Double> votes) {
        double max_vote = -1;
        String winner = null;
        for (String ev : votes.keySet()) {
            if (votes.get(ev) > max_vote) {
                winner = ev;
                max_vote = votes.get(ev);
            }
        }
        return winner;
    }

    /**
     *
     * Obtain weighted_entities, makeinstances, and links
     *
     * @param wl
     * @param weights
     * @param links_support_map
     * @param weighted_entities
     */
    public static void obtainInfoFromMergedLinks(String wl, double[] weights, HashMap<String, String> links_support_map, HashMap<String, String> normmk_normevent_map, HashMap<String, Double>[] weighted_entities, HashMap<String, String> makeinstances, HashMap<String, String> links, String link_threshold) {
        try {
            int linknumber = 1;
            BufferedReader tabreader = new BufferedReader(new FileReader(wl));
            double lower_weight = 0.0;
            //System.out.println("Lower weight ("+lower_weight+")");
            if (link_threshold.equalsIgnoreCase("lower_system_weight")) {
                lower_weight = weights[0];
                for (int i = 1; i < weights.length; i++) {
                    if (weights[i] < lower_weight) {
                        lower_weight = weights[i];
                    }
                }
                lower_weight = lower_weight * (weights.length + 2) + 0.0005;
            } else {
                lower_weight = Double.parseDouble(link_threshold) * (weights.length + 2) + 0.0005;
            }
            System.out.println("Lower weight to omit links ("+lower_weight+")");

            try {
                String line = "";
                while ((line = tabreader.readLine()) != null) {
                    if (line.length() < 1) {
                        continue;
                    }
                    String[] tabsarr = line.split("\t");
                    if (tabsarr.length != 6) {
                        throw new Exception("Malformed merged links file: " + line);
                    }
                   //                     System.out.println(line);

                    // Confidence Threshold (more than weakest system)
                    if (Double.parseDouble(tabsarr[5]) > lower_weight) {
                        //System.err.println("ADDING ("+lower_weight+") "+line);
                        //docid	entity1	entity2	relType	systems weight
                        String[] supportarr = links_support_map.get(tabsarr[1] + "|" + tabsarr[2] + "|" + tabsarr[3]).split(";");
                        if (links_support_map.containsKey(tabsarr[1] + "|" + tabsarr[2] + "|" + tabsarr[3])) {
                            if (tabsarr[1].startsWith("e") && tabsarr[2].startsWith("e")) {
                                if (!makeinstances.containsKey(tabsarr[1])) {
                                    makeinstances.put(tabsarr[1], "<MAKEINSTANCE eiid=\"" + tabsarr[1] + "\" />");
                                }
                                if (!makeinstances.containsKey(tabsarr[2])) {
                                    makeinstances.put(tabsarr[2], "<MAKEINSTANCE eiid=\"" + tabsarr[2] + "\" />");
                                }
                                links.put("l" + linknumber, "<TLINK lid=\"l" + linknumber + "\" eventInstanceID=\"" + tabsarr[1] + "\" relatedToEventInstance=\"" + tabsarr[2] + "\" relType=\"" + tabsarr[3] + "\" />");
                                // weight entities
                                for (int i = 0; i < supportarr.length; i++) {
                                    if (weighted_entities[Integer.parseInt(supportarr[i])].containsKey(normmk_normevent_map.get(tabsarr[1]))) {
                                        weighted_entities[Integer.parseInt(supportarr[i])].put(normmk_normevent_map.get(tabsarr[1]), weighted_entities[Integer.parseInt(supportarr[i])].get(normmk_normevent_map.get(tabsarr[1])) + weights[Integer.parseInt(supportarr[i])]);
                                    } else {
                                        weighted_entities[Integer.parseInt(supportarr[i])].put(normmk_normevent_map.get(tabsarr[1]), weights[Integer.parseInt(supportarr[i])]);
                                    }
                                    if (weighted_entities[Integer.parseInt(supportarr[i])].containsKey(normmk_normevent_map.get(tabsarr[2]))) {
                                        weighted_entities[Integer.parseInt(supportarr[i])].put(normmk_normevent_map.get(tabsarr[2]), weighted_entities[Integer.parseInt(supportarr[i])].get(normmk_normevent_map.get(tabsarr[2])) + weights[Integer.parseInt(supportarr[i])]);
                                    } else {
                                        weighted_entities[Integer.parseInt(supportarr[i])].put(normmk_normevent_map.get(tabsarr[2]), weights[Integer.parseInt(supportarr[i])]);
                                    }
                                }
                            } else {
                                if (tabsarr[1].startsWith("e") && tabsarr[2].startsWith("t")) {
                                    if (!makeinstances.containsKey(tabsarr[1])) {
                                        makeinstances.put(tabsarr[1], "<MAKEINSTANCE eiid=\"" + tabsarr[1] + "\" />");
                                    }
                                    links.put("l" + linknumber, "<TLINK lid=\"l" + linknumber + "\" eventInstanceID=\"" + tabsarr[1] + "\" relatedToTime=\"" + tabsarr[2] + "\" relType=\"" + tabsarr[3] + "\" />");
                                    for (int i = 0; i
                                            < supportarr.length; i++) {
                                        if (weighted_entities[Integer.parseInt(supportarr[i])].containsKey(normmk_normevent_map.get(tabsarr[1]))) {
                                            weighted_entities[Integer.parseInt(supportarr[i])].put(normmk_normevent_map.get(tabsarr[1]), weighted_entities[Integer.parseInt(supportarr[i])].get(normmk_normevent_map.get(tabsarr[1])) + weights[Integer.parseInt(supportarr[i])]);
                                        } else {
                                            weighted_entities[Integer.parseInt(supportarr[i])].put(normmk_normevent_map.get(tabsarr[1]), weights[Integer.parseInt(supportarr[i])]);
                                        }
                                        if (weighted_entities[Integer.parseInt(supportarr[i])].containsKey(tabsarr[2])) {
                                            weighted_entities[Integer.parseInt(supportarr[i])].put(tabsarr[2], weighted_entities[Integer.parseInt(supportarr[i])].get(tabsarr[2]) + weights[Integer.parseInt(supportarr[i])]);
                                        } else {
                                            weighted_entities[Integer.parseInt(supportarr[i])].put(tabsarr[2], weights[Integer.parseInt(supportarr[i])]);
                                        }
                                    }
                                } else {
                                    if (tabsarr[1].startsWith("t") && tabsarr[2].startsWith("t")) {
                                        links.put("l" + linknumber, "<TLINK lid=\"l" + linknumber + "\" timeID=\"" + tabsarr[1] + "\" relatedToTime=\"" + tabsarr[2] + "\" relType=\"" + tabsarr[3] + "\" />");
                                        for (int i = 0; i
                                                < supportarr.length; i++) {
                                            if (weighted_entities[Integer.parseInt(supportarr[i])].containsKey(tabsarr[1])) {
                                                weighted_entities[Integer.parseInt(supportarr[i])].put(tabsarr[1], weighted_entities[Integer.parseInt(supportarr[i])].get(tabsarr[1]) + weights[Integer.parseInt(supportarr[i])]);
                                            } else {
                                                weighted_entities[Integer.parseInt(supportarr[i])].put(tabsarr[1], weights[Integer.parseInt(supportarr[i])]);
                                            }
                                            if (weighted_entities[Integer.parseInt(supportarr[i])].containsKey(tabsarr[2])) {
                                                weighted_entities[Integer.parseInt(supportarr[i])].put(tabsarr[2], weighted_entities[Integer.parseInt(supportarr[i])].get(tabsarr[2]) + weights[Integer.parseInt(supportarr[i])]);
                                            } else {
                                                weighted_entities[Integer.parseInt(supportarr[i])].put(tabsarr[2], weights[Integer.parseInt(supportarr[i])]);
                                            }
                                        }
                                    } else {
                                        throw new Exception("Malformed link: " + line);
                                    }
                                }
                            }
                        } else {
                            throw new Exception("Link in merged that was not supported by systems: " + tabsarr[1] + "|" + tabsarr[2] + "|" + tabsarr[3]);
                        }
                        linknumber++;
                    } else {
                        System.err.println("Skipping (" + lower_weight + ") " + line);
                    }
                }
            } finally {
                if (tabreader == null) {
                    tabreader.close();
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TimeML_Merger):\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
        }
    }

    /**
     *
     * Obtain weighted_entities, makeinstances, and links
     *
     * @param wl
     * @param weights
     * @param links_support_map
     * @param weighted_entities
     */
    public static void obtainInfoFromMergedLinks_bottom_up(String wl, double[] weights, HashMap<String, String> normmk_normevent_map, HashMap<String, String> makeinstances, HashMap<String, String> links, String link_threshold, HashSet<String> merged_entities) {
        try {
            int linknumber = 1;
            BufferedReader tabreader = new BufferedReader(new FileReader(wl));
            double lower_weight = 0.0;
            //System.out.println("Lower weight ("+lower_weight+")");
            if (link_threshold.equalsIgnoreCase("lower_system_weight")) {
                lower_weight = weights[0];
                for (int i = 1; i < weights.length; i++) {
                    if (weights[i] < lower_weight) {
                        lower_weight = weights[i];
                    }
                }
                lower_weight = lower_weight * (weights.length + 2) + 0.0005;
            } else {
                lower_weight = Double.parseDouble(link_threshold) * (weights.length + 2) + 0.0005;
            }
            System.out.println("Lower weight to omit links ("+lower_weight+")");
            try {
                String line = "";
                while ((line = tabreader.readLine()) != null) {
                    if (line.length() < 1) {
                        continue;
                    }
                    String[] tabsarr = line.split("\t");
                    if (tabsarr.length != 6) {
                        throw new Exception("Malformed merged links file: " + line);
                    }
                    //System.out.println(line);
                    // Confidence Threshold (more than weakest system)
                    if (Double.parseDouble(tabsarr[5]) > lower_weight) {
                        //System.err.println("ADDING ("+lower_weight+") "+line);
                        //docid	entity1	entity2	relType	systems weight
                        if (tabsarr[1].startsWith("e") && tabsarr[2].startsWith("e") && merged_entities.contains(normmk_normevent_map.get(tabsarr[1])) && merged_entities.contains(normmk_normevent_map.get(tabsarr[2]))) {
                            if (!makeinstances.containsKey(tabsarr[1])) {
                                makeinstances.put(tabsarr[1], "<MAKEINSTANCE eiid=\"" + tabsarr[1] + "\" />");
                            }
                            if (!makeinstances.containsKey(tabsarr[2])) {
                                makeinstances.put(tabsarr[2], "<MAKEINSTANCE eiid=\"" + tabsarr[2] + "\" />");
                            }
                            links.put("l" + linknumber, "<TLINK lid=\"l" + linknumber + "\" eventInstanceID=\"" + tabsarr[1] + "\" relatedToEventInstance=\"" + tabsarr[2] + "\" relType=\"" + tabsarr[3] + "\" />");
                        } else {
                            if (tabsarr[1].startsWith("e") && tabsarr[2].startsWith("t") && merged_entities.contains(normmk_normevent_map.get(tabsarr[1])) && merged_entities.contains(tabsarr[2])) {
                                if (!makeinstances.containsKey(tabsarr[1])) {
                                    makeinstances.put(tabsarr[1], "<MAKEINSTANCE eiid=\"" + tabsarr[1] + "\" />");
                                }
                                links.put("l" + linknumber, "<TLINK lid=\"l" + linknumber + "\" eventInstanceID=\"" + tabsarr[1] + "\" relatedToTime=\"" + tabsarr[2] + "\" relType=\"" + tabsarr[3] + "\" />");
                            } else {
                                if (tabsarr[1].startsWith("t") && tabsarr[2].startsWith("t") && merged_entities.contains(tabsarr[1]) && merged_entities.contains(tabsarr[2])) {
                                    links.put("l" + linknumber, "<TLINK lid=\"l" + linknumber + "\" timeID=\"" + tabsarr[1] + "\" relatedToTime=\"" + tabsarr[2] + "\" relType=\"" + tabsarr[3] + "\" />");
                                } else {
                                    System.out.println("Malformed link or not supported by merged entities: " + line);
                                }
                            }
                        }
                        linknumber++;
                    } else {
                        System.err.println("Skipping (" + lower_weight + ") " + line);
                    }
                }
            } finally {
                if (tabreader == null) {
                    tabreader.close();
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TimeML_Merger):\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
        }
    }

    /**
     * Write merged output in TimeML format
     * 
     * @param mergedpipes
     * @param makeinstances
     * @param links
     * @param header
     * @param footer
     */
    public static void create_merged_output(String mergedpipes, HashMap<String, String> makeinstances, HashMap<String, String> links, HashSet<String> merged_entities, String header, String footer, String last_text_blanks, String outputfile) {
        BufferedWriter outfile = null;
        int linen = 0;
        try {
            int iob2col = 2;
            int attrscol = iob2col + 1;
            int wordcol = 0;
            BufferedReader pipesreader = new BufferedReader(new FileReader(mergedpipes));
            outfile = new BufferedWriter(new FileWriter(outputfile));
            String inElem = "";
            try {
                if (header == null || header.trim().equals("")) {
                    throw new Exception("Empty header");
                } else {
                    outfile.write(header);
                }
                outfile.write("\n<TEXT>");
                String pipesline;
                String[] pipesarr = null;
                while ((pipesline = pipesreader.readLine()) != null) {
                    linen++;
                    pipesarr = pipesline.split("\\|");
                    String preceding_blanks = "";
                    if (pipesarr[1].matches("[-stn]*")) {
                        int x = pipesarr[1].length();
                        for (int i = 0; i
                                < x; i++) {
                            if (pipesarr[1].charAt(i) == 's') {
                                preceding_blanks += " ";
                            } else {
                                if (pipesarr[1].charAt(i) == 'n') {
                                    preceding_blanks += "\n";
                                } else {
                                    if (pipesarr[1].charAt(i) == 't') {
                                        preceding_blanks += "\t";
                                    }
                                }
                            }
                        }
                    }
                    if (pipesarr[iob2col].startsWith("B")) {
                        if (!inElem.equals("")) {
                            outfile.write("</" + inElem + ">");
                            inElem = "";
                        }
                        String[] attrsarr = pipesarr[attrscol].trim().split(";");
                        inElem = pipesarr[iob2col].substring(2).toUpperCase();
                        outfile.write(preceding_blanks + "<" + inElem);
                        for (int i = 0; i
                                < attrsarr.length; i++) {
                            if (attrsarr[i].indexOf('=') > 0) {
                                String attrname = attrsarr[i].substring(0, attrsarr[i].indexOf('='));
                                String attrvalue = attrsarr[i].substring(attrsarr[i].indexOf("=") + 1);
                                if (attrvalue.matches("\".*\"")) {
                                    attrvalue = attrvalue.substring(1, attrvalue.length() - 1);
                                }
                                if (attrname.matches("(type|value|class|tense|aspect|val|eid|tid|sid)")) {
                                    outfile.write(" " + attrname + "=\"" + attrvalue + "\"");
                                }
                            }
                        }
                        outfile.write(">");
                    }
                    if (pipesarr[iob2col].startsWith("I")) {
                        if (inElem.equals("")) {
                            throw new Exception("Found I-X element without B-X");
                        }
                    }
                    if (pipesarr[iob2col].equals("O")) {
                        if (!inElem.equals("")) {
                            outfile.write("</" + inElem + ">");
                            inElem = "";
                        }
                    }
                    if (!pipesarr[iob2col].startsWith("B")) {
                        outfile.write(preceding_blanks);
                    }
                    outfile.write(pipesarr[wordcol].replaceAll("&#124;", "|").replaceAll("&", "&amp;").replaceAll("<", "&lt;").replaceAll(">", "&gt;"));
                }
                if (!inElem.equals("")) {
                    outfile.write("</" + inElem + ">");
                    inElem = "";
                }
                outfile.write(last_text_blanks + "</TEXT>\n\n");
                // remove links without all entities (were overlaping and one was sacrified)
                // and remove mks out of links
                HashMap<String, String> final_mks = new HashMap<String, String>();
                if (makeinstances != null) {
                    for (String mk : makeinstances.keySet()) {
                        if (makeinstances.get(mk).contains("eventID")) {
                            final_mks.put(mk, makeinstances.get(mk));
                        }
                    }
                }
                makeinstances = (HashMap<String, String>) final_mks.clone();
                final_mks = null;
                final_mks = new HashMap<String, String>();
                HashMap<String, String> final_links = new HashMap<String, String>();
                if (links != null) {
                    for (String lkey : links.keySet()) {
                        String l = links.get(lkey);
                        //System.out.println(l);
                        HashMap<String, String> attribs = XmlAttribs.parseAttrs(l.replaceAll("(<TLINK | />)", ""));
                        if (attribs.containsKey("eventInstanceID") && attribs.containsKey("relatedToEventInstance")) {
                            if (makeinstances.containsKey(attribs.get("eventInstanceID")) && makeinstances.containsKey(attribs.get("relatedToEventInstance"))) {
                                final_mks.put(attribs.get("eventInstanceID"), makeinstances.get(attribs.get("eventInstanceID")));
                                final_mks.put(attribs.get("relatedToEventInstance"), makeinstances.get(attribs.get("relatedToEventInstance")));
                                final_links.put(attribs.get("lid"), l);
                            }
                        } else {
                            if (attribs.containsKey("eventInstanceID") && attribs.containsKey("relatedToTime")) {
                                if (makeinstances.containsKey(attribs.get("eventInstanceID")) && merged_entities.contains(attribs.get("relatedToTime"))) {
                                    final_mks.put(attribs.get("eventInstanceID"), makeinstances.get(attribs.get("eventInstanceID")));
                                    final_links.put(attribs.get("lid"), l);
                                }
                            } else {
                                if (attribs.containsKey("timeID") && attribs.containsKey("relatedToTime")) {
                                    if (merged_entities.contains(attribs.get("timeID")) && merged_entities.contains(attribs.get("relatedToTime"))) {
                                        final_links.put(attribs.get("lid"), l);
                                    }
                                }
                            }
                        }
                    }
                }
                if (final_mks != null) {
                    outfile.write("\n");
                    for (String mk : final_mks.values()) {
                        outfile.write(mk + "\n");
                    }
                    outfile.write("\n");
                }
                if (final_links != null) {
                    outfile.write("\n");
                    List<String> sortedKeys = new ArrayList();
                    sortedKeys.addAll(final_links.keySet());
                    Collections.sort(sortedKeys, new AscINT_lid_Comparator());
                    for (String l : sortedKeys) {
                        outfile.write(final_links.get(l) + "\n");
                    }
                }
                if (footer == null || footer.trim().equals("")) {
                    throw new Exception("Empty footer");
                } else {
                    outfile.write(footer);
                }
            } finally {
                if (pipesreader != null) {
                    pipesreader.close();
                }
                if (outfile != null) {
                    outfile.close();
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TempEval):\n\t" + e.toString() + " (Reading line " + linen + ")\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
        }
    }

    public static String saveWeightedLinks(HashMap<String, Double> weighted_links, HashMap<String, String> names, File f, File output_folder) {
        String outputfile = null;
        try {
            outputfile = output_folder.getCanonicalPath() + File.separator + f.getName() + ".weighted-links";
            BufferedWriter outfile = new BufferedWriter(new FileWriter(outputfile));
            try {
                for (String key : weighted_links.keySet()) {
                    String[] linearr = key.split("\\|");
                    outfile.write(f.getName() + "\t" + linearr[0] + "\t" + linearr[1] + "\t" + linearr[2] + "\t" + names.get(key) + "\t" + weighted_links.get(key) + "\t" + "\n");
                }
            } finally {
                if (outfile != null) {
                    outfile.close();
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TimeML_Merger):\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
            return null;
        }
        return outputfile;
    }

    public static String merge_tok_n_tml(String tokfile, String tmlfile) {
        String outputfile = tokfile + "-IOB2";
        try {
            BufferedWriter outfile = new BufferedWriter(new FileWriter(outputfile));
            boolean hasRoot_tag = false;
            char cxml = '\0';
            String line;
            String tag = "", attribs = "-", inTag = "", inAttribs = "-";
            //boolean closingtag = false;
            char BIO = 'O';
            BufferedReader xmlreader = new BufferedReader(new FileReader(tmlfile));
            BufferedReader pipesreader = new BufferedReader(new FileReader(tokfile));
            try {
                // find root tag
                while (true) {
                    if ((cxml = (char) xmlreader.read()) == -1) {
                        throw new Exception("Premature end of model file");
                    }
                    if (cxml == '<') {
                        if ((cxml = (char) xmlreader.read()) == -1) {
                            throw new Exception("Premature end of model file");
                        }
                        do {
                            tag += cxml;
                            if ((cxml = (char) xmlreader.read()) == -1) {
                                throw new Exception("Premature end of model file");
                            }
                        } while (cxml != '>');
                        if (tag.equalsIgnoreCase("TEXT")) {
                            hasRoot_tag = true;
                            break;
                        }
                        tag = "";
                    } //System.err.print(cxml);
                }
                if (!hasRoot_tag) {
                    throw new Exception("Root tag " + "TEXT" + " not found");
                }
                tag = "";
                cxml = '\0';
                while ((line = pipesreader.readLine()) != null) {
                    if (line.length() < 1) {
                        throw new Exception("Malformed tokens file: empty line.");
                    }
                    String token = line.split("\\|")[0]; // get token
                    boolean interTokenTag = false;
                    boolean findtokenIter = false;
                    boolean delayed_closing = false;
                    char prevxmlchar = 'x';
                    for (int cn = 0; cn < token.length(); cn++) {
                        char cpipes = token.charAt(cn);
                        prevxmlchar = cxml;
                        if ((cxml = (char) xmlreader.read()) == -1) {
                            throw new Exception("Premature end of model file");
                        } //System.err.println("cxml(" + cxml + ") cpipes(" + cpipes + "," + cn + ") "+inTag);
                        if (Character.toLowerCase(cpipes) != Character.toLowerCase(cxml)) {
                            if (cxml == ' ' || cxml == '\n' || cxml == '\r' || cxml == '\t') {
                                cn--;
                                //System.err.println("blank found cn="+cn);
                            } else {
                                // tags handling
                                if (cxml == '<') {
                                    if (cn != 0) {
                                        interTokenTag = true;
                                    }
                                    cn--;
                                    while (((cxml = (char) xmlreader.read()) != (char) -1) && cxml != '>') {
                                        tag += cxml;
                                    }
                                    tag = tag.trim();
                                    if (tag.indexOf(' ') != -1) {
                                        attribs = tag.substring(tag.indexOf(' ') + 1);
                                        tag = tag.substring(0, tag.indexOf(' '));
                                    } //System.err.println("tag=" + tag + " attribs=" + attribs);
                                    if (tag.matches("(?i)(EVENT|TIMEX3)")) {
                                        findtokenIter = true;
                                        //System.err.println("LOOKING opening tag=" + tag + " attribs=" + attribs);
                                        if (interTokenTag) {
                                            System.err.println("Inter-token (" + cn + ") tag consider manual tokenizing: " + token);
                                        }
                                        if (!inTag.equals("")) {
                                            throw new Exception("Nested tags (" + tag + "/" + inTag + ") consider manual correction");
                                        }
                                        inTag = tag;
                                        inAttribs = attribs;
                                        tag = "";
                                        attribs = "-";
                                        BIO = 'B';
                                    } else {
                                        interTokenTag = false;
                                    }
                                    // check if closing
                                    if (tag.matches("/.*")) {
                                        String check = inTag;
                                        if (tag.matches("/" + "(?i)" + check)) {
                                            if (findtokenIter) {
                                                // safe for empty tags (events_4_instances and timex3_4_durations)
                                                if (cn >= 0) {
                                                    System.err.println("Inter Token end of tag (" + inTag + ") cn=" + cn + " " + line);
                                                    delayed_closing = true;
                                                } else {
                                                    BIO = 'O';
                                                    inTag = "";
                                                    inAttribs = "-";
                                                    findtokenIter = false;
                                                    interTokenTag = false;
                                                }
                                            } else {
                                                //System.err.println("closing tag=" + inTag);
                                                BIO = 'O';
                                                inTag = "";
                                            }

                                        }
                                    }
                                    // check if end "TEXT"
                                    if (tag.matches("/" + "(?i)" + "TEXT")) {
                                        System.err.println("closing TEXT=" + "TEXT");
                                        // do something
                                        // it never reaches this because tok file ends before.
                                    }
                                    tag = "";
                                    attribs = "-";
                                } else {
                                    // escaped & < >
                                    if (cxml == '&' || (prevxmlchar == '&' && cxml == 'a')) {
                                        cn--;
                                        while (((cxml = (char) xmlreader.read()) != (char) -1) && cxml != ';') {
                                            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                                                System.err.println("Reading XML escaped char in: " + token);
                                            }
                                        }
                                    } else {
                                        throw new Exception("Distinct chars cxml(" + cxml + ") cpipes(" + cpipes + ")");
                                    }
                                }
                            }
                        }
                    }
                    outfile.write(line + "|" + BIO);
                    if (BIO != 'O') {
                        outfile.write("-" + inTag);
                    }
                    outfile.write("|" + inAttribs + "\n");
                    if (BIO == 'B') {
                        BIO = 'I';
                        inAttribs = "-";
                    }
                    if (delayed_closing) {
                        BIO = 'O';
                        inTag = "";
                        inAttribs = "-";
                        findtokenIter = false;
                        interTokenTag = false;
                        delayed_closing = false;
                    }
                }
            } finally {
                if (pipesreader != null) {
                    pipesreader.close();
                }
                if (xmlreader != null) {
                    xmlreader.close();
                }
                if (outfile != null) {
                    outfile.close();
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TimeML_Merger):\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
            return null;
        }
        return outputfile;
    }

    public static String addSpacing(String tokens, String plainmodel) {
        String outputfile = null;
        int linen = 0;
        try {
            boolean usingNotUTF8tool = false;
            outputfile = tokens + ".spacing";
            BufferedWriter outfile = new BufferedWriter(new FileWriter(outputfile));
            BufferedReader pipesreader = new BufferedReader(new FileReader(tokens));
            BufferedReader modelreader = new BufferedReader(new FileReader(plainmodel));
            try {
                int tokn = 0;
                String pipesline;
                char cmodel = '\0';
                char cmodel_prev = '\0';
                int offset = -1;
                boolean readmodel = true;
                String leadingBlanksString = "-";
                while ((pipesline = pipesreader.readLine()) != null) {
                    linen++;
                    //System.err.println(pipesline);
                    if (pipesline.trim().length() > 0) {
                        String token = pipesline;
                        int token_offset = -1;
                        String paired_token = "";
                        int cn = 0;
                        while (true) {
                            if (readmodel) {
                                cmodel_prev = cmodel;
                                if ((cmodel = (char) modelreader.read()) == -1) {
                                    if (cn == token.length()) {
                                        break; // save last token end of file
                                    } else {
                                        throw new Exception("Premature end of model file");
                                    }
                                }
                                offset++;
                            } else {
                                readmodel = true;
                            }
                            char cpipes = '\0';
                            if (cn >= token.length()) {
                                if (usingNotUTF8tool) {
                                    if (StringUtils.isISO_8859_1(cmodel)) {
                                        readmodel = false;
                                        break;
                                    } else { // delayed token mode for non-ISO desperate cases
                                        cpipes = 'a';
                                    }
                                } else {
                                    readmodel = false;
                                    break;
                                }
                            } else {
                                cpipes = token.charAt(cn);
                            } //System.out.println("offset=" + offset + " cmodel(" + cmodel + ") cpipes(" + cpipes + ")");
                            if (Character.toLowerCase(cpipes) == Character.toLowerCase(cmodel) || (cmodel == '|' && cpipes == '-')) {
                                if (cmodel == '|') {
                                    paired_token += "&#124;"; //scape char for features
                                } else {
                                    paired_token += cmodel;
                                }
                                if (token_offset == -1) {
                                    token_offset = offset;
                                }
                                // multi-dashes problem ('---' is translated by e.g. Roth to '-')
                                if (usingNotUTF8tool && cmodel == '-' && cn == token.length() - 1) {
                                    // read a new char (cmodel) if not end of file to check multi-dash
                                    if (!((cmodel = (char) modelreader.read()) == -1)) {
                                        readmodel = false;
                                        offset++;
                                        if (cmodel == '-') {
                                            cn--;
                                        }
                                    }
                                } else {
                                    readmodel = true;
                                }
                            } else {
                                if (cmodel == ' ' || cmodel == '\t' || cmodel == '\n' || cmodel == '\r') {
                                    cn--;
                                    if (token_offset == -1 && paired_token.equals("")) {
                                        if (cmodel == ' ') {
                                            leadingBlanksString += "s";
                                        } else {
                                            if (cmodel == '\t') {
                                                leadingBlanksString += "t";
                                            } else {
                                                if (cmodel == '\n') {
                                                    leadingBlanksString += "n";
                                                } else {
                                                    if (Character.toLowerCase(cmodel) == '\r') {
                                                        if ((cmodel = (char) modelreader.read()) != (char) -1) {
                                                            offset++;
                                                            if (Character.toLowerCase(cmodel) != '\n') {
                                                                throw new Exception("End of pipesline not found (rn) " + "offset=" + offset + ". cmodel(" + cmodel + ") found instead.");
                                                            } else {
                                                                //DEPRECATED: token_leading_newlines++;
                                                                leadingBlanksString += "n";
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    } else {
                                        // if (cmodel == ' ') {paired_token += "&nbsp;";} // No please
                                        throw new Exception("A space, tab, or newline in the middle of the token cannot be paired, use UTF-8 NLP tools.");
                                    }
                                } else {
                                    // Special for quotes (Roth translates " to `` or '')
                                    if (usingNotUTF8tool && (cmodel == '"' && ((cpipes == '`') || (cpipes == '\'')))) {
                                        if (cn + 1 < token.length() && cpipes == token.charAt(cn + 1)) {
                                            cn += 2;
                                            paired_token += cmodel;
                                        }
                                    } else {
                                        // Special for quotes2 (Roth sudenly changes '' by ``)
                                        if (usingNotUTF8tool && ((cmodel == '\'' || cmodel == '`') && (cpipes == '`' || cpipes == '\''))) {
                                            paired_token += cmodel;
                                        } else {
                                            // multi-dashes problem ('---' is translated by e.g. Roth to '-')
                                            if (usingNotUTF8tool && cmodel == '-' && cmodel_prev == '-') {
                                                paired_token += cmodel;
                                                readmodel = true;
                                                cn--;
                                            } else {
                                                // special for ISO NLP tools
                                                if (usingNotUTF8tool && !StringUtils.isISO_8859_1(cmodel)) {
                                                    paired_token += cmodel;
                                                    readmodel = true;
                                                    cn--;
                                                } else {
                                                    throw new Exception("Distinct chars " + paired_token + " offset=" + offset + " cmodel(" + cmodel + ") cpipes(" + cpipes + ")");
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            cn++;
                        }
                        outfile.write(paired_token + "|" + leadingBlanksString + "\n");
                        leadingBlanksString = "-";
                        tokn++;
                    } else { // newline new sentence
                        tokn = 0;
                    }
                }
            } finally {
                if (pipesreader != null) {
                    pipesreader.close();
                }
                if (modelreader != null) {
                    modelreader.close();
                }
                if (outfile != null) {
                    outfile.close();
                }
            }
        } catch (Exception e) {
            System.err.println("Errors found (TimeML_Merger):\n\t" + e.toString() + "- line:" + linen + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
            return null;
        }
        return outputfile;
    }

    /**
     * Given a relation name, return the inverse Allen-TimeML relation
     * @param rel
     * @return
     */
    public static String reverseRelationCategory(String rel) {
        try {
            if (rel.equals("BEFORE")) {
                return "AFTER";
            }
            if (rel.equals("AFTER")) {
                return "BEFORE";
            }
            if (rel.equals("IBEFORE")) {
                return "IAFTER";
            }
            if (rel.equals("IAFTER")) {
                return "IBEFORE";
            }
            if (rel.equals("DURING")) {
                return "DURING_INV";
            }
            if (rel.equals("BEGINS")) {
                return "BEGUN_BY";
            }
            if (rel.equals("BEGUN_BY")) {
                return "BEGINS";
            }
            if (rel.equals("ENDS")) {
                return "ENDED_BY";
            }
            if (rel.equals("ENDED_BY")) {
                return "ENDS";
            }
            if (rel.equals("OVERLAPS")) {
                return "OVERLAPPED_BY";
            }
            if (rel.equals("OVERLAPPED_BY")) {
                return "OVERLAPS";
            }
            if (rel.equals("INCLUDES")) {
                return "IS_INCLUDED";
            }
            if (rel.equals("IS_INCLUDED")) {
                return "INCLUDES";
            }
            if (rel.equals("IDENTITY") || rel.equals("SIMULTANEOUS")) {
                return "SIMULTANEOUS";
            }
            throw new Exception("Unknow relation: " + rel);
        } catch (Exception e) {
            System.err.println("Errors found:\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
            return null;
        }
    }

    public static String get_last_text_blanks(String file) {
        String last_text_blanks = "";
        try {
            File f = new File(file);
            RandomAccessFile raf = new RandomAccessFile(f, "r");
            //String filecontents = null;
            //filecontents = FileUtils.readFileAsString(file, "UTF-8");
            /*if(filecontents.endsWith("\n")){
            filecontents=filecontents.substring(0, filecontents.length()-1);
            }*/
            /*last_text_blanks = filecontents.replaceAll(".+(\\s+)$", ".$1.");
            if(filecontents.equals(last_text_blanks)){
            last_text_blanks="\n";
            }*/
            for (long i = f.length() - 2; i > 0; i--) {
                raf.seek(i);
                char c = (char) raf.readByte();
                if (c == '\n' || c == '\r' || c == '\t' || c == ' ') {
                    last_text_blanks = c + last_text_blanks;
                } else {
                    break;
                }
            }
            if(raf!=null){
                raf.close();
            }
        } catch (Exception e) {
            System.err.println("Errors found (TML_file_utils):\n\t" + e.toString() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
            }
            return null;
        }
        if (last_text_blanks.equals("")) {
            last_text_blanks = "\n";
        }
        return last_text_blanks;
    }
}
