/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package timemlmerger;

import java.io.*;
import java.util.*;
import nlp_files.XMLFile;
import org.apache.commons.cli.*;
import utils_bk.*;

/**
 *
 * @author hector
 */
public class Main {

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        String annotations = null;
        File output_folder = null;
        String weights = null;
        ArrayList<File[]> annotationList = new ArrayList<File[]>();
        double[] weight=null; // normalized to [0-1]
        double min_et=0.501; // min entity threshol as majority support by default
        String strategy="top-down";
        String link_threshold="lower_system_weight";
        String links_merge_option="consistent";

        try {
            long startTime = System.currentTimeMillis();


            Options opt = new Options();
            //addOption(String opt, boolean hasArg, String description)
            opt.addOption("h", "help", false, "Print this help");
            opt.addOption("a", "annotations", true, "List of folders containing annotations of the same docs (between \"\" and separated by ';')");
            opt.addOption("w", "weights", true, "Optionally annotations can have different float weights (between \"\" and separated by ';', and greater than 0)");
            opt.addOption("o", "output", true, "Output folder to save the merged file/s");
            opt.addOption("et", "entity-threshold", true, "Minimum support for including entities [0-1]");
            opt.addOption("lt", "link-threshold", true, "Minimum support for including links [0-1] (default, autoadjust to min annotation weight)");
            opt.addOption("lmo", "links-merge-option", true, "Links merge option consistent(default) or inconsistent");
            opt.addOption("s", "strategy", true, "top-down (default) or bottom-up");
            opt.addOption("d", "debug", false, "Debug mode: Output errors stack trace (default: disabled)");

            PosixParser parser = new PosixParser();
            CommandLine cl_options = parser.parse(opt, args);
            HelpFormatter hf = new HelpFormatter();
            if (cl_options.hasOption('h')) {
                hf.printHelp("TimeML-Merger", opt);
                System.exit(0);
            } else {
                if (cl_options.hasOption('d')) {
                    System.setProperty("DEBUG", "true");
                }
                String[] annotationsarr = null;
                String[] weightsarr = null;

                if (cl_options.hasOption('a')) {
                    annotations = cl_options.getOptionValue("a");
                    annotationsarr = annotations.split(";");
                    if (annotationsarr.length < 2) {
                        hf.printHelp("TimeML-Merger", opt);
                        throw new Exception("At least TWO annnotations are required.");
                    }
                    for (int i = 0; i < annotationsarr.length; i++) {
                        File f = new File(annotationsarr[i]);
                        if (!f.exists()) {
                            hf.printHelp("TimeML-Merger", opt);
                            throw new Exception("Annotation does not exist: " + annotationsarr[i]);
                        }
                        if (f.isFile()) {
                            File[] files = {f};
                            XMLFile xmlfile = new XMLFile();
                            xmlfile.loadFile(f);
                            xmlfile.overrideExtension("tml-min-consistency");
                            if (!xmlfile.isWellFormed()) {
                                throw new Exception("File: " + xmlfile.getFile().getCanonicalPath() + " is not a valid TimeML XML file.");
                            }
                            annotationList.add(files);
                        } else {
                            File[] files = f.listFiles(FileUtils.onlyFilesFilter);
                            if(files.length==0){
                                throw new Exception("Empty folder: "+f.getName());
                            }
                            for (int fn = 0; fn < files.length; fn++) {
                                XMLFile xmlfile = new XMLFile();
                                xmlfile.loadFile(files[fn]);
                                xmlfile.overrideExtension("tml-min-consistency");
                                if (!xmlfile.isWellFormed()) {
                                    throw new Exception("File: " + xmlfile.getFile().getCanonicalPath() + " is not a valid TimeML XML file.");
                                }
                            }
                            annotationList.add(files);
                        }
                        if (i > 0) {
                            // check equal to previous
                            File[] files1 = annotationList.get(i - 1);
                            File[] files2 = annotationList.get(i);

                            if (files1.length != files2.length) {
                                throw new Exception("Annotation folders must contain exactly the same number of files: "+files1.length);
                            } else {
                                for (int fn = 0; fn < files1.length; fn++) {
                                    if(!files1[fn].getName().equals(files2[fn].getName())){
                                        throw new Exception("Annotation folders must contain exactly the same files: "+files1[fn]);
                                    }
                                }
                            }
                        }
                    }
                } else {
                    hf.printHelp("TimeML-Merger", opt);
                    throw new Exception("Annotations parameter is required.");
                }
                if (cl_options.hasOption("w")) {
                    weights = cl_options.getOptionValue("w");
                    weightsarr = weights.split(";");
                    // WEIGHTS ARE NORMALIZED TO 1.0 for simplicity
                    if (annotationsarr.length != weightsarr.length) {
                        hf.printHelp("TimeML-Merger", opt);
                        throw new Exception("The same number of annotations and weights is required. Found " + annotationsarr.length + " and " + weightsarr.length);
                    } else {
                        weight = new double[annotationList.size()];
                        double sumweight = 0.0;
                        for (int i = 0; i < annotationList.size(); i++) {
                            weight[i] = Double.parseDouble(weightsarr[i]);
                            if(weight[i]<0){
                                throw new Exception("Weirghts must be greater than 0. Found: "+weight[i]);
                            }
                            sumweight+=weight[i];
                        }
                        for (int i = 0; i < annotationList.size(); i++) {
                            weight[i] = weight[i]/sumweight;
                        }
                    }
                } else {
                    // generate equal weights automatically
                    weight = new double[annotationList.size()];
                    double auxweight=1.0/annotationList.size();
                    for (int i = 0; i < annotationList.size(); i++) {
                        weight[i] = auxweight;
                    }
                }

                if (cl_options.hasOption('o')) {
                    output_folder=new File(cl_options.getOptionValue("o"));
                    if(output_folder.exists()){
                        throw new Exception("Output directory already exists: "+output_folder.getCanonicalPath()+". Delete it before runing the program (for safety).");
                    }else{
                        if (!output_folder.mkdirs()) {  // mkdirs creates many parent dirs if needed
                            throw new Exception("Error creating output folder: "+output_folder);
                        }
                    }
                } else {
                    hf.printHelp("TimeML-Merger", opt);
                    throw new Exception("Output folder parameter is required.");
                }

                if (cl_options.hasOption("et")) {
                    min_et=Double.parseDouble(cl_options.getOptionValue("et"));
                    if(min_et>1.0 || min_et<0.0){
                        throw new Exception("Entity threshold must be between 0.0 and 1.0.");
                    }
                }

                if (cl_options.hasOption("et")) {
                    link_threshold=cl_options.getOptionValue("et").toLowerCase();
                }

                if (cl_options.hasOption("s")) {
                    strategy=cl_options.getOptionValue("s").toLowerCase();
                    if(!strategy.matches("(top-down|bottom-up)")){
                        throw new Exception("Strategy must be \"top-down\" or \"bottom-up\".");
                    }
                }

                if (cl_options.hasOption("lmo")) {
                    links_merge_option=cl_options.getOptionValue("lmo").toLowerCase();
                    if(!links_merge_option.matches("(consistent|inconsistent)")){
                        throw new Exception("Links merge option must be \"consistent\" or \"inconsistent\".");
                    }
                }
            }


            System.err.println("Merging with et="+min_et+" and lt="+link_threshold);
            if(strategy.equals("top-down")){
                TimeML_Merger.merge(annotationList,weight,output_folder,min_et,link_threshold, links_merge_option);
            }else{
                TimeML_Merger.merge_bottom_up(annotationList,weight,output_folder,min_et, link_threshold, links_merge_option);
            }

            long endTime = System.currentTimeMillis();
            long sec = (endTime - startTime) / 1000;
            if (sec < 60) {
                System.err.println("Done in " + StringUtils.twoDecPosS(sec) + " sec!\n");
            } else {
                System.err.println("Done in " + StringUtils.twoDecPosS(sec / 60) + " min!\n");
            }
        } catch (Exception e) {
            System.err.println("Errors found:\n\t" + e.getMessage() + "\n");
            if (System.getProperty("DEBUG") != null && System.getProperty("DEBUG").equalsIgnoreCase("true")) {
                e.printStackTrace(System.err);
            }
            System.exit(1);
        }

    }
}
