# this program evaluates systems that extract temporal information from text 
# tlink -> temporal links

#foreach f (24-a-gold-tlinks/data/ABC19980108.1830.0711.tml); do
#python evaluation-relations/code/temporal_evaluation.py $f $(echo $f | p 's/24-a-gold-tlinks/30-b-trips-relations/g')                             
#done

# DURING relations are changed to SIMULTANEOUS

debug = 2

import time 
import sys
import re 
import os

def get_arg (index):
    #for arg in sys.argv:
    if debug >= 3: 
        print sys.argv[index]
    return sys.argv[index]

basedir = re.sub('relation_to_timegraph.py', '', get_arg(0)) 
cmd_folder = os.path.dirname(basedir)
if cmd_folder not in sys.path:
    sys.path.insert(0, cmd_folder)
import relation_to_timegraph

def extract_name(filename):
    parts = re.split('/', filename)
    length = len(parts)
    return parts[length-1]

def get_directory_path(path): 
    name = extract_name(path)
    dir = re.sub(name, '', path) 
    if dir == '': 
        dir = './'
    return dir 


def get_entity_val(word, line): 
    if re.search(word+'="[^"]*"', line): 
        entity = re.findall(word+'="[^"]*"', line)[0]
        entity = re.sub(word+'=', '', entity) 
        entity = re.sub('"', '', entity) 
        return entity 
    return word 
        
def change_DURING_relation(filetext): 
    newtext = '' 
    for line in filetext.split('\n'): 
        foo = '' 
        words = line.split('\t') 
        for i in range(0, len(words)): 
            if i == 3 and words[i] == 'DURING': 
                foo += re.sub('DURING', 'SIMULTANEOUS', words[i]) + '\t'
            else:
                foo += words[i] + '\t' 
        newtext += foo.strip() + '\n' 
    return newtext 

def get_relations(file): 
    text = open(file).read()
    newtext = '' 
    name = extract_name(file) 
    relations = re.findall('<TLINK[^>]*>', text) 
    for each in relations: 
        core = '' 
        ref = '' 
        relType = '' 
        if re.search('eventInstanceID', each): 
            core = get_entity_val('eventInstanceID', each) 
        if re.search('timeID', each): 
            core = get_entity_val('timeID', each) 
        if re.search('relatedToEventInstance', each): 
            ref = get_entity_val('relatedToEventInstance', each) 
        if re.search('relatedToTime', each): 
            ref = get_entity_val('relatedToTime', each) 
        if re.search('relType', each): 
            relType = get_entity_val('relType', each) 
        if core == '' or ref == '' or relType == '': 
            print 'MISSING core, ref or relation', each 
        else: 
            foo = name+'\t'+core+'\t'+ref+'\t'+relType+'\n'
            if debug >= 3: 
                print each 
                print foo 
            newtext += foo + '\n' 
    newtext = change_DURING_relation(newtext)
    return newtext 



def reverse_relation(rel): 
    rel = re.sub('"', '', rel) 
    if rel.upper() == 'BEFORE': 
        return 'AFTER'
    if rel.upper() == 'AFTER': 
        return 'BEFORE' 
    if rel.upper() == 'IBEFORE': 
        return 'IAFTER' 
    if rel.upper() == 'IAFTER': 
        return 'IBEFORE' 
    if rel.upper() == 'DURING': 
        return 'DURING_BY' 
    if rel.upper() == 'BEGINS': 
        return 'BEGUN_BY' 
    if rel.upper() == 'BEGUN_BY': 
        return 'BEGINS'
    if rel.upper() == 'ENDS': 
        return 'ENDED_BY' 
    if rel.upper() == 'ENDED_BY': 
        return 'ENDS' 
    if rel.upper() == 'INCLUDES': 
        return 'IS_INCLUDED' 
    if rel.upper() == 'IS_INCLUDED': 
        return 'INCLUDES' 
    return rel.upper() 


def get_triples(tlink_file): 
    tlinks = tlink_file # open(tlink_file).read() # tlink_file # 
    relations = '' 
    for line in tlinks.split('\n'): 
        if line.strip() == '': 
            continue 
        words = line.split('\t') 
        relations += words[0]+'\t'+words[1]+'\t'+words[2]+'\t'+words[3]+'\n'
        relations += words[0]+'\t'+words[2]+'\t'+words[1]+'\t'+reverse_relation(words[3]) +'\n'        
    return relations 
        
def get_timegraphs(gold, system): 
    gold_text = gold # open(gold).read() # gold #
    system_text = system # open(system).read() # system # 

    tg_gold = relation_to_timegraph.Timegraph() 
    tg_gold = relation_to_timegraph.create_timegraph_from_weight_sorted_relations(gold_text, tg_gold) 
    tg_gold.final_relations = tg_gold.final_relations + tg_gold.violated_relations
    tg_system = relation_to_timegraph.Timegraph() 
    tg_system = relation_to_timegraph.create_timegraph_from_weight_sorted_relations(system_text, tg_system) 
    tg_system.final_relations = tg_system.final_relations + tg_system.violated_relations
    return tg_gold, tg_system 

 
#start_time = time.time()
#end_time = time.time()
 
# extract entities and relation from tlink line 
def get_x_y_rel(tlinks): 
    words = tlinks.split('\t')
    x = words[1]
    y = words[2]
    rel = words[3]
    return x, y, rel 

def get_entity_rel(tlink): 
    words = tlink.split('\t') 
    return words[1]+'\t'+words[2]+'\t'+words[3] 

def total_relation_matched(A_tlinks, B_tlinks, B_relations, B_tg): 
    count = 0 
    for tlink in A_tlinks.split('\n'): 
        if tlink.strip() == '': 
            continue 
        if debug >= 2: 
            print tlink
        x, y, rel = get_x_y_rel(tlink) 
        foo = relation_to_timegraph.interval_rel_X_Y(x, y, B_tg, rel, 'evaluation')
        if re.search(get_entity_rel(tlink.strip()), B_relations): 
            count += 1 
            if debug >= 2: 
                print 'True' 
            continue 
        if debug >= 2: 
            print x, y, rel, foo[1]
        if re.search('true', foo[1]):
            count += 1 
    return count 
            


def evaluate_two_files(arg1, arg2):
    if debug >= 2: 
        print '\n\n Evaluate', arg1, arg2
    gold_annotation = get_relations(arg1)
    system_annotation = get_relations(arg2) 

    tg_gold, tg_system = get_timegraphs(gold_annotation, system_annotation) 
    gold_relations = get_triples(gold_annotation) 
    system_relations = get_triples(system_annotation) 
    
    #for precision
    if debug >= 2: 
        print '\nchecking precision' 
    prec_matched = total_relation_matched(tg_system.final_relations, tg_gold.final_relations, gold_relations, tg_gold) 
    # for recall 
    if debug >= 2: 
        print '\nchecking recall' 
    rec_matched = total_relation_matched(tg_gold.final_relations, tg_system.final_relations, system_relations, tg_system) 

    if debug >= 2: 
        print 'precision', prec_matched, len(tg_system.final_relations.split('\n'))-1
    precision = prec_matched*1.0/(len(tg_system.final_relations.split('\n'))-1)
    if debug >= 2: 
        print 'recall', rec_matched, len(tg_gold.final_relations.split('\n'))-1
    recall = rec_matched*1.0/(len(tg_gold.final_relations.split('\n'))-1)
    print precision, recall 



def evaluate_two_folders(gold, system):
    if gold[-1] != '/': 
        gold += '/' 
    if system[-1] != '/': 
        system += '/' 
    for file in os.listdir(gold):
        if os.path.isdir(gold+file):
            subdir = file+'/'
            if debug >= 1: 
                print 'Traverse files in Directory', gold+subdir
            evaluate_two_folders(gold+subdir, system+subdir)
        else:
            goldfile = gold +  file
            systemfile = system + file 
            if not re.search('DS_Store', file): 
                if debug >= 2: 
                    print goldfile, systemfile 
                evaluate_two_files(goldfile, systemfile)


# take input from command line and give error messages 
# call appropriate functions to evaluate 
def input_and_evaluate(): 
    invalid = 'false' 
    if len(sys.argv) < 3: 
        invalid = 'true' 
    else: 
        arg1 = get_arg(1) 
        arg2 = get_arg(2) 
        global directory_path 
        directory_path = get_directory_path(sys.argv[0])

    # both arguments are directories 
    if invalid == 'false' and os.path.isdir(arg1) and os.path.isdir(arg2): 
        # for each files in gold folder, check the performance of that file in system folder 
        if debug >= 2: 
            print 'compare files in two folders' 
        evaluate_two_folders(arg1, arg2)
    elif invalid == 'false' and os.path.isfile(arg1) and os.path.isfile(arg2): 
        # compare the performance between two files 
        if debug >= 2: 
            print 'compare two files'

        evaluate_two_files(arg1, arg2)
    else: 
        print 'INVALID INPUT FORMAT'
        print '\nto check the performance of a single file:\n\t python evaluate_entities.py gold_file_path system_file_path\n' 
        print 'to check the performace of all files in a gold folder:\n\t python evaluate_entities.py gold_folder_path system_folder_path '
    
    if invalid == 'false': 
        performance = 'get' 
        #get_performance() 



input_and_evaluate()
